from __future__ import annotations

import logging
from enum import Enum

from soda_core.common.data_source_impl import DataSourceImpl
from soda_core.common.data_source_results import QueryResult
from soda_core.common.logging_constants import ExtraKeys, soda_logger
from soda_core.common.sql_dialect import *
from soda_core.contracts.contract_verification import (
    CheckOutcome,
    CheckResult,
    Measurement,
)
from soda_core.contracts.impl.check_types.invalidity_check_yaml import InvalidCheckYaml
from soda_core.contracts.impl.check_types.row_count_check import RowCountMetricImpl
from soda_core.contracts.impl.contract_verification_impl import (
    AggregationMetricImpl,
    CheckImpl,
    CheckParser,
    ColumnImpl,
    ContractImpl,
    DerivedPercentageMetricImpl,
    MeasurementValues,
    MetricImpl,
    MissingAndValidity,
    MissingAndValidityCheckImpl,
    Query,
    ThresholdImpl,
    ThresholdType,
    ValidReferenceData,
)

logger: logging.Logger = soda_logger


class InvalidCheckParser(CheckParser):
    def get_check_type_names(self) -> list[str]:
        return ["invalid"]

    def parse_check(
        self,
        contract_impl: ContractImpl,
        column_impl: Optional[ColumnImpl],
        check_yaml: InvalidCheckYaml,
    ) -> Optional[CheckImpl]:
        return InvalidCheckImpl(
            contract_impl=contract_impl,
            column_impl=column_impl,
            check_yaml=check_yaml,
        )


class InvalidCheckImpl(MissingAndValidityCheckImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        column_impl: ColumnImpl,
        check_yaml: InvalidCheckYaml,
    ):
        super().__init__(
            contract_impl=contract_impl,
            column_impl=column_impl,
            check_yaml=check_yaml,
        )
        self.threshold = ThresholdImpl.create(
            threshold_yaml=check_yaml.threshold,
            default_threshold=ThresholdImpl(type=ThresholdType.SINGLE_COMPARATOR, must_be=0),
        )

        if not self.missing_and_validity.has_validity_configurations():
            logger.error(
                msg="Invalid check does not have any valid or invalid configurations",
                extra={
                    ExtraKeys.LOCATION: self.check_yaml.check_yaml_object.location,
                },
            )

    def setup_metrics(self, contract_impl: ContractImpl, column_impl: ColumnImpl, check_yaml: InvalidCheckYaml):
        self.metric_name = "invalid_percent" if check_yaml.metric == "percent" else "invalid_count"

        self.invalid_count_metric_impl: Optional[MetricImpl] = None
        if self.missing_and_validity.has_reference_data():
            # noinspection PyTypeChecker
            self.invalid_count_metric_impl = self._resolve_metric(
                InvalidReferenceCountMetricImpl(
                    contract_impl=contract_impl, column_impl=column_impl, missing_and_validity=self.missing_and_validity
                )
            )
            # this is used in the check extension to extract failed keys and rows
            self.ref_query = InvalidReferenceCountQuery(
                cte=contract_impl.cte,
                sampler_type=contract_impl.sampler_type,
                sampler_limit=contract_impl.sampler_limit,
                apply_sampling=contract_impl.should_apply_sampling,
                metric_impl=self.invalid_count_metric_impl,
                dataset_filter=self.contract_impl.filter,
                check_filter=self.check_yaml.filter,
                data_source_impl=contract_impl.data_source_impl,
            )
            self.queries.append(self.ref_query)
        else:
            self.invalid_count_metric_impl = self._resolve_metric(
                InvalidCountMetricImpl(contract_impl=contract_impl, column_impl=column_impl, check_impl=self)
            )

        self.row_count_metric = self._resolve_metric(RowCountMetricImpl(contract_impl=contract_impl, check_impl=self))

        self.invalid_percent_metric = self._resolve_metric(
            DerivedPercentageMetricImpl(
                metric_type="invalid_percent",
                fraction_metric_impl=self.invalid_count_metric_impl,
                total_metric_impl=self.row_count_metric,
            )
        )

    def evaluate(self, measurement_values: MeasurementValues) -> CheckResult:
        outcome: CheckOutcome = CheckOutcome.NOT_EVALUATED

        invalid_count: int = measurement_values.get_value(self.invalid_count_metric_impl)
        row_count: int = measurement_values.get_value(self.row_count_metric)
        invalid_percent: float = measurement_values.get_value(self.invalid_percent_metric)

        diagnostic_metric_values: dict[str, float] = {
            "invalid_count": invalid_count,
            "invalid_percent": invalid_percent,
            "check_rows_tested": row_count,
            "dataset_rows_tested": self.contract_impl.dataset_rows_tested,
        }

        threshold_value: Optional[Number] = invalid_percent if self.metric_name == "invalid_percent" else invalid_count

        outcome = self.evaluate_threshold(threshold_value)

        return CheckResult(
            check=self._build_check_info(),
            outcome=outcome,
            threshold_value=threshold_value,
            diagnostic_metric_values=diagnostic_metric_values,
        )

    def get_threshold_metric_impl(self) -> Optional[MetricImpl]:
        return self.invalid_count_metric_impl


class InvalidCountMetricImpl(AggregationMetricImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        column_impl: ColumnImpl,
        check_impl: MissingAndValidityCheckImpl,
    ):
        super().__init__(
            contract_impl=contract_impl,
            column_impl=column_impl,
            metric_type="invalid_count",
            check_filter=check_impl.check_yaml.filter,
            missing_and_validity=check_impl.missing_and_validity,
        )

    def sql_expression(self) -> SqlExpression:
        return SUM(CASE_WHEN(self.sql_condition_expression(), LITERAL(1)))

    def sql_condition_expression(self) -> SqlExpression:
        column_name: str = self.column_impl.column_yaml.name
        return AND.optional(
            [
                SqlExpressionStr.optional(self.check_filter),
                NOT.optional(self.missing_and_validity.is_missing_expr(column_name)),
                self.missing_and_validity.is_invalid_expr(column_name),
            ]
        )

    def convert_db_value(self, value) -> int:
        # Note: expression SUM(CASE WHEN "id" IS NULL THEN 1 ELSE 0 END) gives NULL / None as a result if
        # there are no rows
        return int(value) if value is not None else 0


class InvalidReferenceCountMetricImpl(MetricImpl):
    def __init__(self, contract_impl: ContractImpl, column_impl: ColumnImpl, missing_and_validity: MissingAndValidity):
        super().__init__(
            contract_impl=contract_impl,
            metric_type="invalid_count",
            column_impl=column_impl,
            missing_and_validity=missing_and_validity,
        )


class DatasetAlias(Enum):
    CONTRACT = "C"  # C stands for the 'C'ontract dataset
    REFERENCE = "R"  # R stands for the 'R'eference dataset


class InvalidReferenceCountQuery(Query):
    def __init__(
        self,
        cte: CTE,
        sampler_type: Optional[str],
        sampler_limit: Optional[Number],
        apply_sampling: bool,
        metric_impl: InvalidReferenceCountMetricImpl,
        dataset_filter: Optional[str],
        check_filter: Optional[str],
        data_source_impl: Optional[DataSourceImpl],
    ):
        super().__init__(data_source_impl=data_source_impl, metrics=[metric_impl])
        self.metric_impl = metric_impl
        self.check_filter = check_filter

        self.referencing_alias: str = DatasetAlias.CONTRACT.value
        self.referenced_alias: str = DatasetAlias.REFERENCE.value
        self._referenced_cte_name: str = "_soda_filtered_referenced_dataset"

        self.apply_sampling: bool = apply_sampling
        self.sampler_type: Optional[str] = sampler_type
        self.sampler_limit: Optional[Number] = sampler_limit

        self._cte = cte

        sql_ast = self.build_query(cte=self._cte, select_clause=SELECT(COUNT(STAR())))
        self.sql = self.data_source_impl.sql_dialect.build_select_sql(sql_ast)

    def build_query(self, cte: CTE, select_clause: SqlExpression) -> list[SqlExpression]:
        query = [
            WITH([cte, self.referenced_cte()]),
            select_clause,
            FROM(cte.alias).AS(self.referencing_alias),
            WHERE.optional(SqlExpressionStr.optional(self.check_filter)),
        ]

        query.extend(self.query_join())

        return query

    def referenced_cte(self) -> CTE:
        valid_reference_data: ValidReferenceData = self.metric_impl.missing_and_validity.valid_reference_data
        referenced_dataset_name: str = valid_reference_data.dataset_name
        referenced_dataset_prefix: Optional[list[str]] = valid_reference_data.dataset_prefix

        cte = CTE(self._referenced_cte_name).AS(
            [
                SELECT(STAR()),
                FROM(referenced_dataset_name).IN(referenced_dataset_prefix),
            ]
        )

        if self.apply_sampling and self.sampler_type and self.sampler_limit:
            cte.cte_query[1] = cte.cte_query[1].SAMPLE(self.sampler_type, self.sampler_limit)

        return cte

    def query_join(self) -> SqlExpression:
        valid_reference_data: ValidReferenceData = self.metric_impl.missing_and_validity.valid_reference_data

        referencing_column_name: str = self.metric_impl.column_impl.column_yaml.name

        referenced_dataset_name: str = self._referenced_cte_name
        referenced_column: str = valid_reference_data.column

        # The variant to get the failed rows is:
        # SELECT(STAR().IN("C")),
        # which should translate to SELECT C.*

        is_referencing_column_missing: SqlExpression = self.metric_impl.missing_and_validity.is_missing_expr(
            COLUMN(referencing_column_name).IN(self.referencing_alias)
        )
        is_referencing_column_invalid: SqlExpression = self.metric_impl.missing_and_validity.is_invalid_expr(
            COLUMN(referencing_column_name).IN(self.referencing_alias)
        )
        is_referenced_column_null: SqlExpression = IS_NULL(COLUMN(referenced_column).IN(self.referenced_alias))
        is_referencing_column_invalid: SqlExpression = OR.optional(
            [is_referenced_column_null, is_referencing_column_invalid]
        )

        return [
            LEFT_INNER_JOIN(referenced_dataset_name)
            .ON(
                EQ(
                    COLUMN(referencing_column_name).IN(self.referencing_alias),
                    COLUMN(referenced_column).IN(self.referenced_alias),
                )
            )
            .AS(self.referenced_alias),
            WHERE(AND([NOT(is_referencing_column_missing), is_referencing_column_invalid])),
        ]

    def execute(self) -> list[Measurement]:
        try:
            query_result: QueryResult = self.data_source_impl.execute_query(self.sql)
        except Exception as e:
            logger.error(msg=f"Could not execute invalid reference query {self.sql}: {e}", exc_info=True)
            return []

        metric_value = query_result.rows[0][0]
        metric_impl: MetricImpl = self.metrics[0]
        return [Measurement(metric_id=metric_impl.id, value=metric_value, metric_name=metric_impl.type)]



from __future__ import annotations

from typing import Optional, Type

from soda_core.common.dataset_identifier import DatasetIdentifier
from soda_core.common.soda_cloud import SodaCloud
from soda_core.common.yaml import DataSourceYamlSource


class IContractGenerator:
    _registry: dict[str, Type[IContractGenerator]] = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        fully_qualified_name = cls.__module__ + "." + cls.__name__
        cls._registry[fully_qualified_name] = cls

    @classmethod
    def instance(cls, identifier: Optional[str] = None) -> IContractGenerator:
        if not cls._registry:
            raise NotImplementedError(f"No contract generator implementations registered.")
        if not identifier:
            identifier = next(iter(cls._registry))
        impl_cls = cls._registry[identifier]
        return impl_cls()

    def create_skeleton(
        self,
        data_source_yaml_source: DataSourceYamlSource,
        dataset_identifier: DatasetIdentifier,
        output_file_path: Optional[str],
        verbose: bool,
        soda_cloud: Optional[SodaCloud],
        use_agent: bool,
        generate_checks: bool = True,
    ):
        raise NotImplementedError("No implementation for create_skeleton() yet. Please install an appropriate plugin.")



from __future__ import annotations

import logging

from soda_core.common.data_source_impl import DataSourceImpl
from soda_core.common.data_source_results import QueryResult
from soda_core.common.logging_constants import soda_logger
from soda_core.common.sql_dialect import *
from soda_core.contracts.contract_verification import (
    CheckOutcome,
    CheckResult,
    Measurement,
)
from soda_core.contracts.impl.check_types.metric_check_yaml import MetricCheckYaml
from soda_core.contracts.impl.check_types.row_count_check import RowCountMetricImpl
from soda_core.contracts.impl.contract_verification_impl import (
    AggregationMetricImpl,
    CheckImpl,
    CheckParser,
    ColumnImpl,
    ContractImpl,
    MeasurementValues,
    MetricImpl,
    Query,
    ThresholdImpl,
)

logger: logging.Logger = soda_logger


class MetricCheckParser(CheckParser):
    def get_check_type_names(self) -> list[str]:
        return ["metric"]

    def parse_check(
        self,
        contract_impl: ContractImpl,
        column_impl: Optional[ColumnImpl],
        check_yaml: MetricCheckYaml,
    ) -> Optional[CheckImpl]:
        return MetricCheckImpl(
            contract_impl=contract_impl,
            column_impl=column_impl,
            check_yaml=check_yaml,
        )


class MetricCheckImpl(CheckImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        column_impl: ColumnImpl,
        check_yaml: MetricCheckYaml,
    ):
        super().__init__(
            contract_impl=contract_impl,
            column_impl=column_impl,
            check_yaml=check_yaml,
        )
        self.metric_check_yaml: MetricCheckYaml = check_yaml
        self.threshold = ThresholdImpl.create(
            threshold_yaml=check_yaml.threshold,
        )

    def setup_metrics(
        self,
        contract_impl: ContractImpl,
        column_impl: ColumnImpl,
        check_yaml: MetricCheckYaml,
    ):
        self.numeric_metric_impl: Optional[MetricImpl] = None
        if self.metric_check_yaml.expression:
            self.numeric_metric_impl = self._resolve_metric(
                MetricExpressionMetricImpl(contract_impl=contract_impl, column_impl=column_impl, check_impl=self)
            )
            self.check_rows_tested_metric_impl: MetricImpl = self._resolve_metric(
                RowCountMetricImpl(contract_impl=contract_impl, check_impl=self)
            )

        elif self.metric_check_yaml.query:
            sql = self.metric_check_yaml.query

            if contract_impl.should_apply_sampling:
                sql = contract_impl.data_source_impl.sql_dialect.apply_sampling(
                    sql, contract_impl.sampler_limit, sampler_type=contract_impl.sampler_type
                )

            self.numeric_metric_impl = self._resolve_metric(
                MetricQueryMetricImpl(contract_impl=contract_impl, column_impl=column_impl, check_impl=self)
            )
            if contract_impl.data_source_impl:
                metric_query: Query = MetricQuery(
                    data_source_impl=contract_impl.data_source_impl,
                    metrics=[self.numeric_metric_impl],
                    sql=sql,
                )
                self.queries.append(metric_query)

    def evaluate(self, measurement_values: MeasurementValues) -> CheckResult:
        outcome: CheckOutcome = CheckOutcome.NOT_EVALUATED

        numeric_metric_value: Optional[Number] = measurement_values.get_value(self.numeric_metric_impl)

        diagnostic_metric_values: dict[str, float] = {}

        diagnostic_metric_values: dict[str, float] = {
            "dataset_rows_tested": self.contract_impl.dataset_rows_tested,
        }

        outcome = self.evaluate_threshold(numeric_metric_value)

        return CheckResult(
            check=self._build_check_info(),
            outcome=outcome,
            threshold_value=numeric_metric_value,
            diagnostic_metric_values=diagnostic_metric_values,
        )


class MetricExpressionMetricImpl(AggregationMetricImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        column_impl: ColumnImpl,
        check_impl: MetricCheckImpl,
        expression: str | None = None,
        data_source_impl: Optional[DataSourceImpl] = None,
        dataset_identifier: Optional[DatasetIdentifier] = None,
    ):
        self.expression: str = expression or check_impl.check_yaml.expression
        super().__init__(
            contract_impl=contract_impl,
            column_impl=column_impl,
            metric_type=check_impl.type,
            check_filter=check_impl.check_yaml.filter,
            data_source_impl=data_source_impl,
            dataset_identifier=dataset_identifier,
        )

    def _get_id_properties(self) -> dict[str, any]:
        id_properties: dict[str, any] = super()._get_id_properties()
        id_properties["expression"] = self.expression
        return id_properties

    def sql_expression(self) -> SqlExpression:
        return SqlExpressionStr(self.expression)

    def convert_db_value(self, value) -> any:
        return float(value) if value is not None else None


class MetricQueryMetricImpl(MetricImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        column_impl: ColumnImpl,
        check_impl: MetricCheckImpl,
        query: Optional[str] = None,
        data_source_impl: Optional[DataSourceImpl] = None,
        dataset_identifier: Optional[DatasetIdentifier] = None,
    ):
        self.query: str = query or check_impl.check_yaml.query
        super().__init__(
            contract_impl=contract_impl,
            column_impl=column_impl,
            metric_type=check_impl.type,
            check_filter=check_impl.check_yaml.filter,
            data_source_impl=data_source_impl,
            dataset_identifier=dataset_identifier,
        )

    def _get_id_properties(self) -> dict[str, any]:
        id_properties: dict[str, any] = super()._get_id_properties()
        id_properties["query"] = self.query
        return id_properties


class MetricQuery(Query):
    def __init__(self, data_source_impl: Optional[DataSourceImpl], metrics: list[MetricImpl], sql: str):
        super().__init__(data_source_impl=data_source_impl, metrics=metrics, sql=sql)

    def build_sql(self) -> str:
        return super().build_sql()

    def execute(self) -> list[Measurement]:
        try:
            query_result: QueryResult = self.data_source_impl.execute_query(self.sql)
        except Exception as e:
            logger.error(msg=f"Could not execute metric query: \n{self.sql}:\n{e}", exc_info=True)
            return []

        metric_value = query_result.rows[0][0]
        metric_impl: MetricImpl = self.metrics[0]
        return [Measurement(metric_id=metric_impl.id, value=metric_value, metric_name=metric_impl.type)]



from __future__ import annotations

from numbers import Number

import sqlglot
from soda_core.common.sql_ast import SamplerType
from sqlglot import exp


def build_sample_clause(sampler_limit: Number, sampler_type: SamplerType) -> exp.TableSample:
    if sampler_limit <= 0:
        raise ValueError("sampler_limit must be positive")

    size = exp.Literal.number(sampler_limit)
    sample = exp.TableSample()

    if sampler_type == SamplerType.ABSOLUTE_LIMIT:
        sample.set("size", size)
    else:
        raise ValueError(f"Unsupported sample type: {sampler_type}")

    return sample


def attach_sample_to_relation(rel: exp.Expression, sampler_limit: Number, sampler_type: SamplerType) -> None:
    """
    Attach a TableSample clause to a relation (Table or Subquery),
    unless it already has one.
    """
    if rel is None:
        return

    if rel.args.get("sample"):
        return

    if isinstance(rel, (exp.Table, exp.Subquery)):
        rel.set("sample", build_sample_clause(sampler_limit, sampler_type))


def apply_sampling_to_sql(
    sql: str,
    sampler_limit: Number,
    sampler_type: SamplerType,
    read_dialect: str | None = None,
    write_dialect: str | None = None,
) -> str:
    """
    Add TABLESAMPLE / SAMPLE to every table-like source in all FROM and JOIN clauses,
    including inside CTEs and subqueries.

    Exact rendering is dialect-specific.
    """
    tree = sqlglot.parse_one(sql, read=read_dialect) if read_dialect else sqlglot.parse_one(sql)

    # FROM sources (top-level, CTE bodies, nested subqueries)
    # Keep track of CTEs and skip them as they are already sampled at their definition
    # Skip subqueries as they are sampled at their definition. We skip the FROM SUBQUERY part, the FROM within the subquery is handled when we process that subquery separately.
    ctes = {cte.alias_or_name for cte in tree.find_all(exp.CTE)}
    for from_ in tree.find_all(exp.From):
        if isinstance(from_.this, exp.Table) and from_.this.alias_or_name in ctes:
            continue
        if isinstance(from_.this, exp.Subquery):
            continue

        attach_sample_to_relation(from_.this, sampler_limit, sampler_type)

    # JOIN targets
    for join in tree.find_all(exp.Join):
        attach_sample_to_relation(join.this, sampler_limit, sampler_type)

    return tree.sql(dialect=write_dialect, pretty=True) if write_dialect else tree.sql(pretty=True)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum


# Create a enum of table types
class TableType(Enum):
    TABLE = "BASE TABLE"
    VIEW = "VIEW"
    MATERIALIZED_VIEW = "MATERIALIZED VIEW"


@dataclass
class FullyQualifiedObjectName:
    database_name: str
    schema_name: str

    def get_object_name(self) -> str:
        # Returns the specific object name (e.g. the table name, or the view name)
        raise NotImplementedError("Subclasses must implement this method")


@dataclass
class FullyQualifiedTableName(FullyQualifiedObjectName):
    table_name: str

    def get_object_name(self) -> str:
        return self.table_name


@dataclass
class FullyQualifiedViewName(FullyQualifiedObjectName):
    view_name: str

    def get_object_name(self) -> str:
        return self.view_name


@dataclass
class FullyQualifiedMaterializedViewName(FullyQualifiedObjectName):
    materialized_view_name: str

    def get_object_name(self) -> str:
        return self.materialized_view_name



from os.path import dirname, exists
from pathlib import Path
from typing import Dict, Optional

from soda_core.cli.exit_codes import ExitCode
from soda_core.common.exceptions import (
    ContractParserException,
    InvalidArgumentException,
    InvalidDataSourceConfigurationException,
)
from soda_core.common.logging_constants import Emoticons, soda_logger
from soda_core.common.yaml import ContractYamlSource
from soda_core.contracts.api import test_contract, verify_contract
from soda_core.contracts.api.publish_api import publish_contract
from soda_core.contracts.contract_verification import ContractVerificationSessionResult


def handle_verify_contract(
    contract_file_path: Optional[str],
    dataset_identifier: Optional[str],
    data_source_file_paths: Optional[str],
    soda_cloud_file_path: Optional[str],
    variables: Optional[Dict[str, str]],
    publish: bool,
    verbose: bool,
    use_agent: bool,
    blocking_timeout_in_minutes: int,
    check_paths: list[str],
    diagnostics_warehouse_file_path: Optional[str],
) -> ExitCode:
    try:
        contract_verification_result = verify_contract(
            contract_file_path=contract_file_path,
            dataset_identifier=dataset_identifier,
            data_source_file_path=None,
            data_source_file_paths=data_source_file_paths,
            soda_cloud_file_path=soda_cloud_file_path,
            variables=variables,
            publish=publish,
            verbose=verbose,
            use_agent=use_agent,
            blocking_timeout_in_minutes=blocking_timeout_in_minutes,
            check_paths=check_paths,
            dwh_data_source_file_path=diagnostics_warehouse_file_path,
        )

        return interpret_contract_verification_result(contract_verification_result)

    except (InvalidArgumentException, InvalidDataSourceConfigurationException, Exception) as exc:
        soda_logger.exception(f"An unexpected exception occurred: {exc}")
        return ExitCode.LOG_ERRORS


def contract_verification_is_not_sent_to_cloud(
    contract_verification_session_result: ContractVerificationSessionResult,
) -> bool:
    return any(
        cr.sending_results_to_soda_cloud_failed
        for cr in contract_verification_session_result.contract_verification_results
    )


def interpret_contract_verification_result(verification_result: ContractVerificationSessionResult) -> ExitCode:
    if contract_verification_is_not_sent_to_cloud(verification_result):
        return ExitCode.RESULTS_NOT_SENT_TO_CLOUD
    elif verification_result.has_errors:
        return ExitCode.LOG_ERRORS
    elif verification_result.is_failed:
        return ExitCode.CHECK_FAILURES
    elif verification_result.is_warned:
        return ExitCode.CHECK_WARNINGS
    else:
        return ExitCode.OK


def handle_publish_contract(
    contract_file_path: Optional[str],
    soda_cloud_file_path: Optional[str],
) -> ExitCode:
    try:
        contract_publication_result = publish_contract(contract_file_path, soda_cloud_file_path)
        if contract_publication_result.has_errors:
            # TODO: detect/deal with exit code 4?
            return ExitCode.LOG_ERRORS
        else:
            return ExitCode.OK
    except ContractParserException as exc:
        soda_logger.error(f"Failed to parse contract: {exc}")
        return ExitCode.LOG_ERRORS
    except Exception as exc:
        soda_logger.exception(f"Failed to publish contract: {exc}")
        return ExitCode.LOG_ERRORS


def handle_test_contract(
    contract_file_path: Optional[str],
    variables: Optional[Dict[str, str]],
) -> ExitCode:
    contract_verification_result = test_contract(contract_file_path=contract_file_path, variables=variables)
    if contract_verification_result.has_errors:
        return ExitCode.LOG_ERRORS
    else:
        soda_logger.info(f"{Emoticons.WHITE_CHECK_MARK} {contract_file_path} is valid")
        return ExitCode.OK


def handle_fetch_contract(
    contract_file_paths: Optional[list[str]],
    dataset_identifiers: Optional[list[str]],
    soda_cloud_file_path: Optional[str],
) -> ExitCode:
    from soda_core.common.soda_cloud import SodaCloud

    soda_cloud_client: Optional[SodaCloud] = None
    try:
        if soda_cloud_file_path:
            soda_cloud_client = SodaCloud.from_config(soda_cloud_file_path)

        validate_fetch_arguments(contract_file_paths, dataset_identifiers, soda_cloud_client)

        # fetch contract YAML strings
        contract_yaml_sources = []
        if soda_cloud_client:
            for dataset_identifier in dataset_identifiers:
                contract = soda_cloud_client.fetch_contract_for_dataset(dataset_identifier)
                if not contract:
                    soda_logger.error(f"Could not fetch contract for dataset '{dataset_identifier}': skipping fetch")
                    continue
                contract_yaml_sources.append(ContractYamlSource.from_str(contract))
        if not contract_yaml_sources or len(contract_yaml_sources) == 0:
            soda_logger.debug("No contracts given. Exiting.")
            return ExitCode.OK

        # write to local contract YAML files
        for contract_file_path, contract_yaml in zip(contract_file_paths, contract_yaml_sources):
            dir_name = dirname(contract_file_path)
            try:
                Path(dir_name).mkdir(parents=True, exist_ok=True)
                if exists(contract_file_path):
                    action = "Updated"
                else:
                    action = "Created"
                with open(contract_file_path, "w") as contract_file:
                    contract_file.write(contract_yaml.yaml_str)
                soda_logger.info(f"{Emoticons.WHITE_CHECK_MARK}  {action} contract source file '{contract_file_path}'")
                return ExitCode.OK
            except Exception as exc:
                soda_logger.error(f"An unexpected exception occurred: {exc}")
                return ExitCode.LOG_ERRORS

    except (InvalidArgumentException, Exception) as exc:
        soda_logger.exception(f"Unexpected error during contract fetching: {exc}")
        return ExitCode.LOG_ERRORS


def validate_fetch_arguments(
    contract_file_paths: Optional[list[str]],
    dataset_identifiers: Optional[list[str]],
    soda_cloud_file_path: Optional[str],
) -> None:
    if not contract_file_paths:
        raise InvalidArgumentException(
            "A Soda Data Contract file path is required to use the fetch command. "
            "Please provide the '-f/--file' argument with a valid contract file path."
        )
    if not dataset_identifiers:
        raise InvalidArgumentException(
            "A Soda Cloud dataset identifier is required to use the fetch command. "
            "Please provide the '-d/--dataset' argument with a valid dataset identifier."
        )
    if not soda_cloud_file_path:
        raise InvalidArgumentException(
            "A Soda Cloud configuration file is required to use the fetch command. "
            "Please provide the '--soda-cloud' argument with a valid configuration file path."
        )
    if len(contract_file_paths) != len(dataset_identifiers):
        raise InvalidArgumentException(
            "The number of contract file paths must match the number of dataset identifiers. "
            "Ensure that each contract corresponds to a specific dataset."
        )



from __future__ import annotations

import logging
from typing import Optional

from soda_core.common.logging_constants import ExtraKeys, soda_logger
from soda_core.common.yaml import YamlObject
from soda_core.contracts.impl.contract_yaml import (
    CheckYaml,
    CheckYamlParser,
    ColumnYaml,
    MissingAncValidityCheckYaml,
)

logger: logging.Logger = soda_logger


class FreshnessCheckYamlParser(CheckYamlParser):
    def get_check_type_names(self) -> list[str]:
        return ["freshness"]

    def parse_check_yaml(
        self, check_type_name: str, check_yaml_object: YamlObject, column_yaml: Optional[ColumnYaml]
    ) -> Optional[CheckYaml]:
        if column_yaml:
            logger.error(
                msg=f"The 'freshness' check type is not supported in column checks. Was on column '{column_yaml.name}'",
                extra={ExtraKeys.LOCATION: check_yaml_object.location},
            )
        return FreshnessCheckYaml(type_name=check_type_name, check_yaml_object=check_yaml_object)


class FreshnessCheckYaml(MissingAncValidityCheckYaml):
    def __init__(self, type_name: str, check_yaml_object: YamlObject):
        super().__init__(type_name=type_name, check_yaml_object=check_yaml_object)
        self.column: str = check_yaml_object.read_string("column")
        self.now_variable: Optional[str] = check_yaml_object.read_string_opt("now_variable")
        if self.unit and not self.unit in ["minute", "hour", "day"]:
            logger.error(
                msg=f"Invalid freshness threshold unit: {self.unit}",
                extra={ExtraKeys.LOCATION: check_yaml_object.create_location_from_yaml_dict_key("unit")},
            )
            self.unit = None

    def get_valid_units(self) -> list[str]:
        return ["minute", "hour", "day"]



from __future__ import annotations

from typing import Optional

from soda_core.common.yaml import YamlObject
from soda_core.contracts.impl.contract_yaml import (
    CheckYaml,
    CheckYamlParser,
    ColumnYaml,
    MissingAncValidityCheckYaml,
)


class MissingCheckYamlParser(CheckYamlParser):
    def get_check_type_names(self) -> list[str]:
        return ["missing"]

    def parse_check_yaml(
        self, check_type_name: str, check_yaml_object: YamlObject, column_yaml: Optional[ColumnYaml]
    ) -> Optional[CheckYaml]:
        return MissingCheckYaml(type_name=check_type_name, check_yaml_object=check_yaml_object)


class MissingCheckYaml(MissingAncValidityCheckYaml):
    def __init__(self, type_name: str, check_yaml_object: YamlObject):
        super().__init__(type_name=type_name, check_yaml_object=check_yaml_object)



from __future__ import annotations

import ast
import inspect
import logging
import textwrap
from functools import wraps
from typing import Optional

from opentelemetry import trace
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator
from opentelemetry.trace.span import Span
from opentelemetry.trace.status import Status, StatusCode
from soda_core.cli.exit_codes import OK_CODES, ExitCode

trace_context_propagator = TraceContextTextMapPropagator()
trace_context_carrier = {}

tracer = trace.get_tracer_provider().get_tracer(__name__)


from soda_core.common.logging_constants import soda_logger

logger: logging.Logger = soda_logger


def get_decorators(function):
    decorators = {}

    def visit_function_def(node):
        decorators[node.name] = {}
        for n in node.decorator_list:
            name = ""
            if isinstance(n, ast.Call):
                group = n.func.value.id
                name = n.func.attr if isinstance(n.func, ast.Attribute) else n.func.id

                for a in n.args:
                    print(ast.dump(a))
            else:
                group = n.attr if isinstance(n, ast.Attribute) else n.id
                name = None

            if group not in decorators[node.name]:
                decorators[node.name][group] = []

            if name:
                decorators[node.name][group].append({name})

    node_iter = ast.NodeVisitor()
    node_iter.visit_function_def = visit_function_def
    node_iter.visit(ast.parse(textwrap.dedent(inspect.getsource(function))))
    return decorators


def soda_trace(fn: callable):
    def _after_exec(span: Span, error: Optional[BaseException] = None):
        if error:
            code = getattr(error, "code", None)
            try:
                if ExitCode(code) in OK_CODES:
                    span.set_status(Status(StatusCode.OK))
                else:
                    span.set_status(Status(StatusCode.ERROR, str(error)))
            except (ValueError, TypeError):
                span.set_status(Status(StatusCode.ERROR, str(error)))
        else:
            span.set_status(Status(StatusCode.OK))

    @wraps(fn)
    def wrapper(*original_args, **original_kwargs):
        ctx = trace_context_propagator.extract(carrier=trace_context_carrier)
        with tracer.start_as_current_span(f"{fn.__module__}.{fn.__name__}", context=ctx) as span:
            trace_context_propagator.inject(carrier=trace_context_carrier)
            result = None
            try:
                result = fn(*original_args, **original_kwargs)
                _after_exec(span)
            except BaseException as e:
                # Catch base exception to deal with system exit codes as well.
                _after_exec(span, e)
                raise e
        return result

    return wrapper



from __future__ import annotations

import logging
from abc import abstractmethod
from datetime import date, datetime, time
from numbers import Number
from textwrap import indent
from typing import TYPE_CHECKING, Any, Optional, Tuple

from soda_core.common.data_source_results import QueryResult
from soda_core.common.dataset_identifier import DatasetIdentifier
from soda_core.common.datetime_conversions import (
    convert_datetime_to_str,
    convert_str_to_datetime,
)
from soda_core.common.logging_constants import soda_logger
from soda_core.common.metadata_types import (
    ColumnMetadata,
    DataSourceNamespace,
    SamplerType,
    SodaDataTypeName,
    SqlDataType,
)
from soda_core.common.sql_ast import (
    ALTER_TABLE,
    ALTER_TABLE_ADD_COLUMN,
    ALTER_TABLE_DROP_COLUMN,
    AND,
    AVERAGE,
    CASE_WHEN,
    CAST,
    COALESCE,
    COLUMN,
    COMBINED_HASH,
    CONCAT,
    CONCAT_WS,
    COUNT,
    CREATE_MATERIALIZED_VIEW,
    CREATE_TABLE,
    CREATE_TABLE_AS_SELECT,
    CREATE_TABLE_COLUMN,
    CREATE_TABLE_IF_NOT_EXISTS,
    CREATE_VIEW,
    CTE,
    DISTINCT,
    DROP_MATERIALIZED_VIEW,
    DROP_MATERIALIZED_VIEW_IF_EXISTS,
    DROP_TABLE,
    DROP_TABLE_IF_EXISTS,
    DROP_VIEW,
    DROP_VIEW_IF_EXISTS,
    EQ,
    EXISTS,
    FROM,
    FUNCTION,
    GROUP_BY,
    GT,
    GTE,
    IN,
    IN_SELECT,
    INSERT_INTO,
    INSERT_INTO_VIA_SELECT,
    INTO,
    IS_NOT_NULL,
    IS_NULL,
    JOIN,
    LEFT_INNER_JOIN,
    LENGTH,
    LIKE,
    LIMIT,
    LITERAL,
    LOWER,
    LT,
    LTE,
    MAX,
    MIN,
    NEQ,
    NOT,
    NOT_LIKE,
    OFFSET,
    OR,
    ORDER_BY_ASC,
    ORDER_BY_DESC,
    ORDINAL_POSITION,
    RAW_SQL,
    REGEX_LIKE,
    SELECT,
    STAR,
    STRING_HASH,
    SUM,
    TUPLE,
    UNION,
    UNION_ALL,
    VALUES,
    VALUES_ROW,
    WHERE,
    WITH,
    Operator,
    SqlExpression,
    SqlExpressionStr,
)
from soda_core.common.sql_utils import apply_sampling_to_sql
from soda_core.common.statements.table_types import FullyQualifiedObjectName, TableType
from typing_extensions import deprecated

if TYPE_CHECKING:
    from soda_core.common.data_source_impl import DataSourceImpl

logger: logging.Logger = soda_logger


class SqlDialect:
    """
    Extends DataSource with all logic to builds the SQL queries.
    Specific DataSource's can customize their SQL queries by subclassing SqlDialect,
    overriding methods of SqlDialect and returning the customized SqlDialect in DataSource._create_sql_dialect()
    """

    DEFAULT_QUOTE_CHAR = '"'

    SODA_DATA_TYPE_SYNONYMS: tuple[tuple[SodaDataTypeName, ...]] = ()

    def __init__(
        self,
        data_source_impl: DataSourceImpl,
    ):
        self.data_source_impl: DataSourceImpl = data_source_impl

        self._data_type_name_synonym_mappings: dict[str, str] = self._build_data_type_name_synonym_mappings(
            self._get_data_type_name_synonyms()
        )

    # Data type handling

    def _build_data_type_name_synonym_mappings(self, data_type_name_synonyms: list[list[str]]) -> dict[str, str]:
        data_type_name_synonym_mappings: dict[str, str] = {}
        for data_type_name_synonym_list in data_type_name_synonyms:
            first_type_lower: str = data_type_name_synonym_list[0].lower()
            for data_type_name_synonym in data_type_name_synonym_list:
                data_type_name_synonym_mappings[data_type_name_synonym.lower()] = first_type_lower
        return data_type_name_synonym_mappings

    def _get_data_type_name_synonyms(self) -> list[list[str]]:
        # Implements data type synonyms
        # Each list should represent a list of synonyms
        return [
            # Eg for postgres
            # ["varchar", "character varying"],
            # ["char", "character"],
            # ["integer", "int", "int4"],
            # ["bigint", "int8"],
            # ["smallint", "int2"],
            # ["real", "float4"],
            # ["double precision", "float8"],
        ]

    def data_type_names_are_same_or_synonym(self, left_data_type_name: str, right_data_type_name: str) -> bool:
        left_data_type_name_lower: str = left_data_type_name.lower()
        right_data_type_name_lower: str = right_data_type_name.lower()
        if left_data_type_name_lower == right_data_type_name_lower:
            return True
        left_synonym_data_type_name: str = self._data_type_name_synonym_mappings.get(
            left_data_type_name_lower, left_data_type_name_lower
        )
        right_synonym_data_type_name: str = self._data_type_name_synonym_mappings.get(
            right_data_type_name_lower, right_data_type_name_lower
        )
        return left_synonym_data_type_name == right_synonym_data_type_name

    def get_data_source_data_type_name_for_soda_data_type_name(self, soda_data_type_name: SodaDataTypeName) -> str:
        return self.get_data_source_data_type_name_by_soda_data_type_names()[soda_data_type_name]

    @abstractmethod
    def get_data_source_data_type_name_by_soda_data_type_names(self) -> dict[SodaDataTypeName, str]:
        """
        Maps SodaDataTypeName's names to native data source type names.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_soda_data_type_name_by_data_source_data_type_names(self) -> dict[str, SodaDataTypeName]:
        """
        Maps native data source type names to SodaDataTypeName's
        """
        raise NotImplementedError()

    def is_same_data_type_for_dwh_column(self, expected: SqlDataType, actual: SqlDataType):
        return self.is_same_data_type_for_schema_check(expected=expected, actual=actual)

    def is_same_data_type_for_schema_check(self, expected: SqlDataType, actual: SqlDataType):
        if not self.data_type_names_are_same_or_synonym(expected.name, actual.name):
            return False
        if (
            isinstance(expected.character_maximum_length, int)
            and expected.character_maximum_length != actual.character_maximum_length
        ):
            return False
        if isinstance(expected.numeric_precision, int) and expected.numeric_precision != actual.numeric_precision:
            return False
        if isinstance(expected.numeric_scale, int) and expected.numeric_scale != actual.numeric_scale:
            return False
        if isinstance(expected.datetime_precision, int) and expected.datetime_precision != actual.datetime_precision:
            return False
        return True

    def get_synonyms_for_soda_data_type(self) -> list[list[SodaDataTypeName]]:
        # This function can be overloaded if required.
        # It could be that the datasource has synonyms for the data types, and we want to handle that in the mappings.
        # For an example: see the postgres implementation. We basically create a list of lists. Whereby each sublist contains the synonyms for a given SodaDataTypeName.
        return []

    @classmethod
    def is_same_soda_data_type_with_synonyms(cls, expected: SodaDataTypeName, actual: SodaDataTypeName) -> bool:
        if expected == actual:
            return True

        for synonyms in cls.SODA_DATA_TYPE_SYNONYMS:
            if expected in synonyms and actual in synonyms:
                logger.debug(
                    f"In is_same_soda_data_type_with_synonyms, expected {expected} and actual {actual} are treated as the same"
                )
                return True

        return False

    def map_test_sql_data_type_to_data_source(self, source_data_type: SqlDataType) -> SqlDataType:
        test_data_type: str = source_data_type.name
        data_type_name: str = self.get_data_source_data_type_name_by_soda_data_type_names().get(test_data_type)
        character_maximum_length: Optional[int] = (
            source_data_type.character_maximum_length if self.supports_data_type_character_maximum_length() else None
        )
        numeric_precision: Optional[int] = (
            source_data_type.numeric_precision if self.supports_data_type_numeric_precision() else None
        )
        numeric_scale: Optional[int] = (
            source_data_type.numeric_scale if self.supports_data_type_numeric_scale() else None
        )
        datetime_precision: Optional[int] = (
            source_data_type.datetime_precision if self.supports_data_type_datetime_precision() else None
        )
        return self.get_sql_data_type_class()(
            name=data_type_name,
            character_maximum_length=character_maximum_length,
            numeric_precision=numeric_precision,
            numeric_scale=numeric_scale,
            datetime_precision=datetime_precision,
        )

    def data_type_has_parameter_character_maximum_length(self, data_type_name) -> bool:
        return self.format_metadata_data_type(data_type_name).lower() in [
            "varchar",
            "char",
            "character varying",
            "character",
        ]

    def data_type_has_parameter_numeric_precision(self, data_type_name) -> bool:
        return self.format_metadata_data_type(data_type_name).lower() in ["numeric", "number", "decimal"]

    def data_type_has_parameter_numeric_scale(self, data_type_name) -> bool:
        return self.format_metadata_data_type(data_type_name).lower() in ["numeric", "number", "decimal"]

    def data_type_has_parameter_datetime_precision(self, data_type_name) -> bool:
        return self.format_metadata_data_type(data_type_name).lower() in [
            "timestamp",
            "timestamp without time zone",
            "timestamptz",
            "timestamp with time zone",
            "time",
            "time with time zone",
        ]

    # SQL generation

    def quote_default(self, identifier: Optional[str]) -> Optional[str]:
        return (
            f"{self.DEFAULT_QUOTE_CHAR}{identifier}{self.DEFAULT_QUOTE_CHAR}"
            if isinstance(identifier, str) and len(identifier) > 0
            else None
        )

    def build_fully_qualified_sql_name(self, dataset_identifier: DatasetIdentifier) -> str:
        return self.qualify_dataset_name(
            dataset_prefix=dataset_identifier.prefixes, dataset_name=dataset_identifier.dataset_name
        )

    def qualify_dataset_name(self, dataset_prefix: list[str], dataset_name: str) -> str:
        """
        Creates a fully qualified table name, optionally quoting the table name
        """
        parts: list[str] = list(dataset_prefix) if dataset_prefix else []
        parts.append(dataset_name)
        parts = [self.quote_default(p) for p in parts if p]
        return ".".join(parts)

    def literal(self, o: object) -> str:
        if o is None:
            return "NULL"
        elif isinstance(o, Number) and not isinstance(
            o, bool
        ):  # We don't want to interpret booleans as numbers. In Python a boolean is a subclass of int, so we have to explicitly check it
            return self.literal_number(o)
        elif isinstance(o, str):
            return self.literal_string(o)
        elif isinstance(o, datetime):
            if o.tzinfo is None:
                return self.literal_datetime(o)
            else:
                return self.literal_datetime_with_tz(o)
        elif isinstance(o, time):
            return self.literal_time(o)
        elif isinstance(o, date):
            return self.literal_date(o)
        elif isinstance(o, list) or isinstance(o, set) or isinstance(o, tuple):
            return self.literal_list(o)
        elif isinstance(o, bool):
            return self.literal_boolean(o)
        elif isinstance(o, LITERAL):  # If someone passes a LITERAL object, we want to use the value
            return self.literal(o.value)
        raise RuntimeError(f"Cannot convert type {type(o)} to a SQL literal: {o}")

    def literal_number(self, value: Number):
        if value is None:
            return None
        return str(value)

    def literal_string(self, value: str) -> Optional[str]:
        if value is None:
            return None
        return "'" + self.escape_string(value) + "'"

    def literal_list(self, l: list):
        if l is None:
            return None
        return "(" + (",".join([self.literal(e) for e in l])) + ")"

    def literal_date(self, date: date):
        date_string = date.strftime("%Y-%m-%d")
        return f"DATE '{date_string}'"

    def literal_datetime(self, datetime: datetime):
        return f"'{datetime.isoformat()}'"

    def literal_time(self, time: time):
        time_str: str = time.strftime("%H:%M:%S.%f")
        return f"'{time_str}'"

    def literal_datetime_with_tz(self, datetime: datetime):
        # Can be overloaded if the subclass does not support timezones (may have to do conversion yourself)
        # We assume that all timestamps are stored in UTC.
        # See Fabric for an example
        return self.literal_datetime(datetime)

    def literal_boolean(self, boolean: bool):
        return "TRUE" if boolean is True else "FALSE"

    def escape_string(self, value: str):
        # string_literal: str = re.sub(r"(\\.)", r"\\\1", value)
        string_literal: str = value.replace("'", "''")
        return string_literal

    def escape_regex(self, value: str):
        return value

    def create_schema_if_not_exists_sql(self, prefixes: list[str], add_semicolon: bool = True) -> str:
        assert len(prefixes) == 2, f"Expected 2 prefixes, got {len(prefixes)}"
        schema_name: str = prefixes[1]
        quoted_schema_name: str = self.quote_default(schema_name)
        return f"CREATE SCHEMA IF NOT EXISTS {quoted_schema_name}" + (";" if add_semicolon else "")

    def post_schema_create_sql(self, prefixes: list[str]) -> Optional[list[str]]:
        """Optional list of SQL commands to execute after schema is created (e.g. to set permissions)"""
        return None

    def select_all_paginated_sql(
        self,
        dataset_identifier: DatasetIdentifier,
        columns: list[str],
        filter: Optional[str],
        order_by: list[str],
        limit: int,
        offset: int,
    ) -> str:
        where_clauses = []

        if filter:
            where_clauses.append(SqlExpressionStr(filter))

        statements = [
            SELECT(columns or [STAR()]),
            FROM(table_name=dataset_identifier.dataset_name, table_prefix=dataset_identifier.prefixes),
            WHERE.optional(AND.optional(where_clauses)),
            *[ORDER_BY_ASC(c) for c in order_by],
            LIMIT(limit),
            OFFSET(offset),
        ]

        return self.build_select_sql(statements)

    #########################################################
    # CREATE TABLE
    #########################################################
    def build_create_table_sql(
        self, create_table: CREATE_TABLE | CREATE_TABLE_IF_NOT_EXISTS, add_semicolon: bool = True
    ) -> str:
        create_table_sql = self._build_create_table_statement_sql(create_table)

        create_table_sql = (
            create_table_sql
            + "(\n"
            + ",\n".join([self._build_create_table_column(column) for column in create_table.columns])
            + "\n)"
        )
        return create_table_sql + (";" if add_semicolon else "")

    def _build_create_table_statement_sql(self, create_table: CREATE_TABLE | CREATE_TABLE_IF_NOT_EXISTS) -> str:
        if_not_exists_sql: str = "IF NOT EXISTS" if isinstance(create_table, CREATE_TABLE_IF_NOT_EXISTS) else ""
        create_table_sql: str = f"CREATE TABLE {if_not_exists_sql} {create_table.fully_qualified_table_name} "
        return create_table_sql

    def _build_create_table_column(self, create_table_column: CREATE_TABLE_COLUMN) -> str:
        column_name_quoted: str = self._quote_column_for_create_table(create_table_column.name)
        column_type_sql: str = self._build_create_table_column_type(create_table_column)

        is_nullable_sql: str = (
            " NOT NULL" if (create_table_column.nullable is False and self._is_not_null_ddl_supported()) else ""
        )
        default_sql: str = (
            f" DEFAULT {self.literal(create_table_column.default)}" if create_table_column.default else ""
        )

        return f"\t{column_name_quoted} {column_type_sql}{is_nullable_sql}{default_sql}"

    def _build_create_table_column_type(self, create_table_column: CREATE_TABLE_COLUMN) -> str:
        assert isinstance(create_table_column.type, SqlDataType)

        if not self.supports_data_type_character_maximum_length():
            create_table_column.type.character_maximum_length = None
        if not self.supports_data_type_numeric_precision():
            create_table_column.type.numeric_precision = None
        if not self.supports_data_type_numeric_scale():
            create_table_column.type.numeric_scale = None
        if not self.supports_data_type_datetime_precision():
            create_table_column.type.datetime_precision = None

        return create_table_column.type.get_sql_data_type_str_with_parameters()

    def _quote_column_for_create_table(self, column_name: str) -> str:
        return self.quote_default(
            column_name
        )  # Some datasources (Athena) require a different quoting when creating a table.

    def _is_not_null_ddl_supported(self) -> bool:
        return True

    def build_create_table_as_select_sql(
        self, create_table_as_select: CREATE_TABLE_AS_SELECT, add_semicolon: bool = True, add_parenthesis: bool = True
    ) -> str:
        pre_parenthesis_sql: str = "(" if add_parenthesis else ""
        post_parenthesis_sql: str = ")" if add_parenthesis else ""
        result_sql: str = f"CREATE TABLE {create_table_as_select.fully_qualified_table_name} AS "
        result_sql += (
            f"{pre_parenthesis_sql}\n{self.build_select_sql(create_table_as_select.select_elements, add_semicolon=False)}{post_parenthesis_sql}"
            + (";" if add_semicolon else "")
        )
        return result_sql

    #########################################################
    # ALTER TABLE
    #########################################################
    def build_alter_table_sql(self, alter_table: ALTER_TABLE, add_semicolon: bool = True) -> str:
        if isinstance(alter_table, ALTER_TABLE_ADD_COLUMN):
            return self._build_alter_table_add_column_sql(alter_table, add_semicolon)
        elif isinstance(alter_table, ALTER_TABLE_DROP_COLUMN):
            return self._build_alter_table_drop_column_sql(alter_table, add_semicolon)
        else:
            raise ValueError(f"Unexpected alter table type: {alter_table.__class__.__name__}")

    def _build_alter_table_add_column_sql(
        self, alter_table: ALTER_TABLE_ADD_COLUMN, add_semicolon: bool = True, add_parenthesis: bool = False
    ) -> str:
        column_name_quoted: str = self._quote_column_for_create_table(alter_table.column.name)
        column_type_sql: str = self._build_create_table_column_type(alter_table.column)
        is_nullable_sql: str = (
            " NOT NULL" if (alter_table.column.nullable is False and self._is_not_null_ddl_supported()) else ""
        )
        default_sql: str = f" DEFAULT {self.literal(alter_table.column.default)}" if alter_table.column.default else ""
        pre_parenthesis_sql: str = "(" if add_parenthesis else ""
        post_parenthesis_sql: str = ")" if add_parenthesis else ""
        return (
            f"ALTER TABLE {alter_table.fully_qualified_table_name} {self._get_add_column_sql_expr()} {pre_parenthesis_sql}{column_name_quoted} {column_type_sql}{is_nullable_sql}{default_sql}{post_parenthesis_sql}"
            + (";" if add_semicolon else "")
        )

    def _get_add_column_sql_expr(self) -> str:
        return "ADD COLUMN"

    def _build_alter_table_drop_column_sql(
        self, alter_table: ALTER_TABLE_DROP_COLUMN, add_semicolon: bool = True
    ) -> str:
        column_name_quoted: str = self._quote_column_for_create_table(alter_table.column_name)
        return f"ALTER TABLE {alter_table.fully_qualified_table_name} DROP COLUMN {column_name_quoted}" + (
            ";" if add_semicolon else ""
        )

    def drop_column_supported(self) -> bool:
        return True

    #########################################################
    # DROP TABLE
    #########################################################
    def build_drop_table_sql(self, drop_table: DROP_TABLE | DROP_TABLE_IF_EXISTS, add_semicolon: bool = True) -> str:
        if_exists_sql: str = "IF EXISTS " if isinstance(drop_table, DROP_TABLE_IF_EXISTS) else ""
        cascade_sql: str = " CASCADE" if getattr(drop_table, "cascade", False) else ""
        return f"DROP TABLE {if_exists_sql}{drop_table.fully_qualified_table_name}{cascade_sql}" + (
            ";" if add_semicolon else ""
        )

    #########################################################
    # INSERT INTO
    #########################################################
    def build_insert_into_sql(self, insert_into: INSERT_INTO, add_semicolon: bool = True) -> str:
        insert_into_sql: str = f"INSERT INTO {insert_into.fully_qualified_table_name}"
        insert_into_sql += self._build_insert_into_columns_sql(insert_into)
        insert_into_sql += self._build_insert_into_values_sql(insert_into)
        return insert_into_sql + (";" if add_semicolon else "")

    def _build_insert_into_columns_sql(self, insert_into: INSERT_INTO) -> str:
        columns_sql: str = " (" + ", ".join([self.build_expression_sql(column) for column in insert_into.columns]) + ")"
        return columns_sql

    def build_insert_into_via_select_sql(
        self, insert_into_via_select: INSERT_INTO_VIA_SELECT, add_semicolon: bool = True
    ) -> str:
        insert_into_sql: str = f"INSERT INTO {insert_into_via_select.fully_qualified_table_name}\n"
        insert_into_sql += self._build_insert_into_columns_sql(insert_into_via_select) + "\n"
        insert_into_sql += (
            "(\n" + self.build_select_sql(insert_into_via_select.select_elements, add_semicolon=False) + "\n)"
        )
        return insert_into_sql + (";" if add_semicolon else "")

    def _build_insert_into_values_sql(self, insert_into: INSERT_INTO) -> str:
        values_sql: str = " VALUES\n" + ",\n".join(
            [self._build_insert_into_values_row_sql(value) for value in insert_into.values]
        )
        return values_sql

    def build_cte_values_sql(self, values: VALUES, alias_columns: list[COLUMN] | None) -> str:
        return " VALUES\n" + ",\n".join([self.build_expression_sql(value) for value in values.values])

    def _build_insert_into_values_row_sql(self, values: VALUES_ROW) -> str:
        values_sql: str = "(" + ", ".join([self.literal(value) for value in values.values]) + ")"
        # We used to have this line here to escape special characters.
        # But this doesn't accurately insert text fields with escaped characters (such as newlines): values_sql = self.encode_string_for_sql(values_sql)
        return values_sql

    #########################################################
    # CREATE VIEW
    #########################################################
    def build_create_view_sql(
        self, create_view: CREATE_VIEW, add_semicolon: bool = True, add_parenthesis: bool = True
    ) -> str:
        pre_parenthesis_sql: str = "(" if add_parenthesis else ""
        post_parenthesis_sql: str = ")" if add_parenthesis else ""
        select_sql: str = self.build_select_sql(create_view.select_elements, add_semicolon=False)
        return (
            f"CREATE VIEW {create_view.fully_qualified_view_name} AS {pre_parenthesis_sql}\n{select_sql}{post_parenthesis_sql}\n"
            + (";" if add_semicolon else "")
        )

    def build_drop_view_sql(self, drop_view: DROP_VIEW | DROP_VIEW_IF_EXISTS, add_semicolon: bool = True) -> str:
        if_exists_sql: str = "IF EXISTS " if isinstance(drop_view, DROP_VIEW_IF_EXISTS) else ""
        return f"DROP VIEW {if_exists_sql}{drop_view.fully_qualified_view_name}" + (";" if add_semicolon else "")

    #########################################################
    # CREATE MATERIALIZED VIEW
    #########################################################
    def build_create_materialized_view_sql(
        self,
        create_materialized_view: CREATE_MATERIALIZED_VIEW,
        add_semicolon: bool = True,
        add_parenthesis: bool = True,
    ) -> str:
        pre_parenthesis_sql: str = "(" if add_parenthesis else ""
        post_parenthesis_sql: str = ")" if add_parenthesis else ""
        select_sql: str = self.build_select_sql(create_materialized_view.select_elements, add_semicolon=False)
        return (
            f"CREATE MATERIALIZED VIEW {create_materialized_view.fully_qualified_view_name} AS {pre_parenthesis_sql}\n{select_sql}{post_parenthesis_sql}\n"
            + (";" if add_semicolon else "")
        )

    def build_drop_materialized_view_sql(
        self, drop_view: DROP_MATERIALIZED_VIEW | DROP_MATERIALIZED_VIEW_IF_EXISTS, add_semicolon: bool = True
    ) -> str:
        if_exists_sql: str = "IF EXISTS " if isinstance(drop_view, DROP_MATERIALIZED_VIEW_IF_EXISTS) else ""
        return f"DROP MATERIALIZED VIEW {if_exists_sql}{drop_view.fully_qualified_view_name}" + (
            ";" if add_semicolon else ""
        )

    #########################################################
    # UNION
    #########################################################
    @deprecated(
        "Not fully supported for all datasources yet. Only tested/verified for Postgres and Redshift. Use with caution."
    )
    def build_union_sql(self, union: UNION | UNION_ALL, add_semicolon: bool = True) -> str:
        # TODO: add support for other datasources. Currently only tested/verified for Postgres.
        union_sql: str = "UNION ALL" if isinstance(union, UNION_ALL) else "UNION"
        return f"\n{union_sql}\n".join(
            [
                f"(\n{self.build_select_sql(select_element, add_semicolon=False)}\n)"
                for select_element in union.select_elements
            ]
        ) + (";" if add_semicolon else "")

    #########################################################
    # SELECT
    #########################################################

    # TODO: refactor this to use AST (`SELECT`) instead of a list of `select_elements`. See inherited overriden methods as well.
    def build_select_sql(self, select_elements: list, add_semicolon: bool = True) -> str:
        statement_lines: list[str] = []
        statement_lines.extend(self._build_cte_sql_lines(select_elements))
        statement_lines.extend(self._build_select_sql_lines(select_elements))
        statement_lines.extend(self._build_into_sql_lines(select_elements))
        statement_lines.extend(self._build_from_sql_lines(select_elements))
        statement_lines.extend(self._build_where_sql_lines(select_elements))
        statement_lines.extend(self._build_group_by_sql_lines(select_elements))
        statement_lines.extend(self._build_order_by_lines(select_elements))

        limit_line = self._build_limit_line(select_elements)
        if limit_line:
            statement_lines.append(limit_line)

        offset_line = self._build_offset_line(select_elements)
        if offset_line:
            statement_lines.append(offset_line)
        return "\n".join(statement_lines) + (";" if add_semicolon else "")

    def _build_select_sql_lines(self, select_elements: list) -> list[str]:
        select_field_sqls: list[str] = []
        for select_element in select_elements:
            if isinstance(select_element, SELECT):
                if isinstance(select_element.fields, str) or isinstance(select_element.fields, SqlExpression):
                    select_element.fields = [select_element.fields]
                for select_field in select_element.fields:
                    if isinstance(select_field, str):
                        select_field_sqls.append(self.quote_default(select_field))
                    elif isinstance(select_field, SqlExpression):
                        select_field_sqls.append(self.build_expression_sql(select_field))
                    else:
                        raise Exception(f"Invalid select field type: {select_field.__class__.__name__}")

        # Alternatively, concatenate all the fields on one line to reduce SQL statement length
        # return "SELECT " + (", ".join(select_fields_sql))
        # For now, we opt for SELECT statement readability...

        select_sql_lines: list[str] = []
        for i in range(0, len(select_field_sqls)):
            if i == 0:
                sql_line = f"SELECT {select_field_sqls[0]}"
            else:
                sql_line = f"       {select_field_sqls[i]}"
            # Append comma all lines except the last one
            if i < len(select_field_sqls) - 1:
                sql_line += ","
            select_sql_lines.append(sql_line)

        return select_sql_lines

    def _build_into_sql_lines(self, select_elements: list) -> list[str]:
        into_lines: list[str] = []
        for select_element in select_elements:
            if isinstance(select_element, INTO):
                into_lines.append(f"INTO {select_element.fully_qualified_table_name}")
        return into_lines

    def _build_cte(self, cte: CTE) -> str:
        if isinstance(cte.cte_query, list):
            sql_str: str = self.build_select_sql(cte.cte_query)
        elif isinstance(cte.cte_query, VALUES):
            sql_str: str = self.build_cte_values_sql(values=cte.cte_query, alias_columns=cte.alias_columns)
        elif isinstance(cte.cte_query, str):
            sql_str: str = indent(cte.cte_query, "  ").strip()
        else:
            raise ValueError(f"Unexpected cte_query type: {cte.cte_query.__class__.__name__}")
        sql_str = sql_str.rstrip(";")
        # Do not indent the query string! This could cause issues with multi-line text fields. sql_str = indent(sql_str, "  ")
        alias_columns_str: str = ""
        if cte.alias_columns and self.supports_cte_alias_columns():
            alias_columns_str = "(" + ", ".join([self._build_column_sql(column) for column in cte.alias_columns]) + ")"
        return f"{self.quote_default(cte.alias)}{alias_columns_str} AS (\n{sql_str}\n)"

    def supports_cte_alias_columns(self) -> bool:
        return True

    def _build_cte_sql_lines(self, select_elements: list) -> list[str]:
        cte_lines: list[str] = []
        for select_element in select_elements:
            if isinstance(select_element, WITH):
                cte_sql: list[str] = [self._build_cte(cte) for cte in select_element.cte_list]
                cte_sql = "WITH \n" + ",\n".join(cte_sql)
                cte_lines = cte_sql.split("\n")
        return cte_lines

    def build_expression_sql(self, expression: SqlExpression | str | Number) -> str:
        if isinstance(expression, str):
            return self.quote_default(expression)
        elif isinstance(expression, Number):
            return str(expression)
        elif isinstance(expression, COLUMN):
            return self._build_column_sql(expression)
        elif isinstance(expression, LITERAL):
            return self.literal(expression.value)
        elif isinstance(expression, RAW_SQL):
            return expression.value
        elif isinstance(expression, OR):
            return self._build_or_sql(expression)
        elif isinstance(expression, AND):
            return self._build_and_sql(expression)
        elif isinstance(expression, NOT):
            return self._build_not_sql(expression)
        elif isinstance(expression, Operator):
            return self._build_operator_sql(expression)
        elif isinstance(expression, COUNT):
            return self._build_count_sql(expression)
        elif isinstance(expression, SUM):
            return self._build_sum_sql(expression)
        elif isinstance(expression, CASE_WHEN):
            return self._build_case_when_sql(expression)
        elif isinstance(expression, TUPLE):
            return self._build_tuple_sql(expression)
        elif isinstance(expression, COMBINED_HASH):
            return self._build_combined_hash_sql(expression)
        elif isinstance(expression, CONCAT):
            return self._build_concat_sql(expression)
        elif isinstance(expression, CONCAT_WS):
            return self._build_concat_ws_sql(expression)
        elif isinstance(expression, STRING_HASH):
            return self._build_string_hash_sql(expression)
        elif isinstance(expression, IS_NULL):
            return self._build_is_null_sql(expression)
        elif isinstance(expression, IS_NOT_NULL):
            return self._build_is_not_null_sql(expression)
        elif isinstance(expression, REGEX_LIKE):
            return self._build_regex_like_sql(expression)
        elif isinstance(expression, LIKE):
            return self._build_like_sql(expression)
        elif isinstance(expression, IN):
            return self._build_in_sql(expression)
        elif isinstance(expression, IN_SELECT):
            return self._build_in_select_sql(expression)
        elif isinstance(expression, NOT_LIKE):
            return self._build_not_like_sql(expression)
        elif isinstance(expression, LOWER):
            return self._build_lower_sql(expression)
        elif isinstance(expression, LENGTH):
            return self._build_length_sql(expression)
        elif isinstance(expression, MAX):
            return self._build_max_sql(expression)
        elif isinstance(expression, MIN):
            return self._build_min_sql(expression)
        elif isinstance(expression, AVERAGE):
            return self._build_average_sql(expression)
        elif isinstance(expression, COALESCE):
            return self._build_coalesce_sql(expression)
        elif isinstance(expression, CAST):
            return self._build_cast_sql(expression)
        elif isinstance(expression, FUNCTION):
            return self._build_function_sql(expression)
        elif isinstance(expression, DISTINCT):
            return self._build_distinct_sql(expression)
        elif isinstance(expression, SqlExpressionStr):
            return f"({expression.expression_str})"
        elif isinstance(expression, ORDINAL_POSITION):
            return self._build_ordinal_position_sql(expression)
        elif isinstance(expression, STAR):
            return self._build_star_sql(expression)
        elif isinstance(expression, EXISTS):
            return self._build_exists_sql(expression)
        raise Exception(f"Invalid expression type {expression.__class__.__name__}")

    def _build_column_sql(self, column: COLUMN) -> str:
        table_alias_sql: str = f"{self.quote_default(column.table_alias)}." if column.table_alias else ""
        column_sql: str = self.build_expression_sql(
            column.name
        )  # If column.name is a SqlExpression, it will be compiled; if a string, it will be quoted
        field_alias_sql: str = f" AS {self.quote_default(column.field_alias)}" if column.field_alias else ""
        return f"{table_alias_sql}{column_sql}{field_alias_sql}"

    def _build_or_sql(self, or_expr: OR) -> str:
        if isinstance(or_expr.clauses, list) and len(or_expr.clauses) == 1:
            return self.build_expression_sql(or_expr.clauses[0])
        if isinstance(or_expr.clauses, str) or isinstance(or_expr.clauses, SqlExpression):
            return self.build_expression_sql(or_expr.clauses)
        or_clauses_sql: str = " OR ".join(self.build_expression_sql(or_clause) for or_clause in or_expr.clauses)
        return f"({or_clauses_sql})"

    def _build_not_sql(self, not_expr: NOT) -> str:
        expr_sql: str = self.build_expression_sql(not_expr.expression)
        return f"NOT({expr_sql})"

    def _build_and_sql(self, and_expr: AND) -> str:
        if isinstance(and_expr.clauses, list) and len(and_expr.clauses) == 1:
            return self.build_expression_sql(and_expr.clauses[0])
        if isinstance(and_expr.clauses, str) or isinstance(and_expr.clauses, SqlExpression):
            return self.build_expression_sql(and_expr.clauses)
        return " AND ".join(self.build_expression_sql(and_clause) for and_clause in and_expr.clauses)

    def _build_from_sql_lines(self, select_elements: list) -> list[str]:
        sql_lines: list[str] = []
        # This method formats with newlines and indentation.
        # Alternatively, concatenate all the fields on one line to reduce SQL statement length
        # return "SELECT " + (", ".join(select_fields_sql))
        # For now, we opt for SELECT statement readability...

        from_elements: list[FROM] = [
            select_element for select_element in select_elements if isinstance(select_element, FROM)
        ]

        from_sql_line: str = "FROM "
        for from_element in from_elements:
            if type(from_element) == FROM:
                if from_element is not from_elements[0]:
                    sql_lines.append(f"{from_sql_line},")
                    from_sql_line = "     "
                from_sql_line += self._build_from_part(from_element)
            elif isinstance(from_element, LEFT_INNER_JOIN) or isinstance(from_element, JOIN):
                sql_lines.append(from_sql_line)
                from_sql_line = f"     {self._build_join_part(from_element)}"

        sql_lines.append(from_sql_line)
        return sql_lines

    def get_from_name_from_qualified_name(self, qualified_name: str) -> str:
        # FROM in the current AST will always quote the table name. So we need to check it, and remove the first and last quote characters if they exist.
        return qualified_name[1:-1] if self.is_quoted(qualified_name) else qualified_name

    def convert_fully_qualified_object_name_to_sql_qualified_name(
        self, fully_qualified_object_name: FullyQualifiedObjectName
    ) -> str:
        return ".".join(
            [
                self.quote_default(name_part)
                for name_part in [
                    fully_qualified_object_name.database_name,
                    fully_qualified_object_name.schema_name,
                    fully_qualified_object_name.get_object_name(),
                ]
                if isinstance(name_part, str)
            ]
        )

    def _alias_format(self, alias: str) -> str:
        return f"AS {self.quote_default(alias)}"

    def _build_from_part(self, from_part: FROM) -> str:
        # "fully".qualified"."tablename" [AS "table_alias"]

        from_parts: list[str] = [
            self._build_qualified_quoted_dataset_name(
                dataset_name=from_part.table_name, dataset_prefix=from_part.table_prefix
            )
        ]

        if from_part.sampler_type is not None and isinstance(from_part.sample_size, Number):
            from_parts.append(self._build_sample_sql(from_part.sampler_type, from_part.sample_size))

        if isinstance(from_part.alias, str):
            from_parts.append(self._alias_format(from_part.alias))

        return " ".join(from_parts)

    def _build_join_part(self, join: LEFT_INNER_JOIN | JOIN) -> str:
        # [INNER JOIN] "fully".qualified"."tablename" [AS "table_alias"] [ON join_condition]

        from_parts: list[str] = []

        if isinstance(join, LEFT_INNER_JOIN):
            from_parts.append("LEFT JOIN")
        if isinstance(join, JOIN):
            from_parts.append("JOIN")

        from_parts.append(
            self._build_qualified_quoted_dataset_name(dataset_name=join.table_name, dataset_prefix=join.table_prefix)
        )

        if isinstance(join.alias, str):
            from_parts.append(self._alias_format(join.alias))

        if isinstance(join, LEFT_INNER_JOIN) or isinstance(join, JOIN):
            from_parts.append(f"ON {self.build_expression_sql(join.on_condition)}")

        return " ".join(from_parts)

    def _build_qualified_quoted_dataset_name(self, dataset_name: str, dataset_prefix: Optional[list[str]]) -> str:
        name_parts: list[str] = [] if dataset_prefix is None else list(dataset_prefix)
        name_parts.append(dataset_name)
        quoted_name_parts: list[str] = [
            self.quote_default(name_part) for name_part in name_parts if isinstance(name_part, str)
        ]
        return ".".join(quoted_name_parts)

    def _build_operator_sql(self, operator: Operator) -> str:
        operators: dict[type, str] = {
            EQ: "=",
            NEQ: "!=",
            LT: "<",
            LTE: "<=",
            GT: ">",
            GTE: ">=",
            LIKE: "like",
        }
        operator_sql: str = operators[type(operator)]
        return f"{self.build_expression_sql(operator.left)} {operator_sql} {self.build_expression_sql(operator.right)}"

    def _build_where_sql_lines(self, select_elements: list) -> list[str]:
        and_expressions: list[SqlExpression] = []
        for select_element in select_elements:
            if isinstance(select_element, WHERE):
                and_expressions.append(select_element.condition)
            elif isinstance(select_element, AND):
                and_expressions.extend(select_element._get_clauses_as_list())

        where_parts: list[str] = [self.build_expression_sql(and_expression) for and_expression in and_expressions]

        where_sql_lines: list[str] = []
        for i in range(0, len(where_parts)):
            if i == 0:
                sql_line = f"WHERE {where_parts[0]}"
            else:
                sql_line = f"  AND {where_parts[i]}"
            where_sql_lines.append(sql_line)
        return where_sql_lines

    def _build_group_by_sql_lines(self, select_elements: list) -> list[str]:
        group_by_field_sqls: list[str] = []
        for select_element in select_elements:
            if isinstance(select_element, GROUP_BY):
                if isinstance(select_element.fields, str) or isinstance(select_element.fields, SqlExpression):
                    select_element.fields = [select_element.fields]
                group_by_field_sqls.extend(
                    [self.build_expression_sql(select_field) for select_field in select_element.fields]
                )
        sql_lines: list[str] = []
        if group_by_field_sqls:
            group_by_fields_str: str = ", ".join(group_by_field_sqls)
            sql_lines.append(f"GROUP BY {group_by_fields_str}")
        return sql_lines

    def _build_function_sql(self, function: FUNCTION) -> str:
        args: list[SqlExpression | str] = [function.args] if not isinstance(function.args, list) else function.args
        args_sqls: list[str] = [self.build_expression_sql(arg) for arg in args]
        if function.name in ["+", "-", "/", "*"]:
            operators: str = f" {function.name} ".join(args_sqls)
            return f"({operators})"
        else:
            args_list_sql: str = ", ".join(args_sqls)
            return f"{function.name}({args_list_sql})"

    def _build_star_sql(self, star: STAR) -> str:
        if star.alias:
            return f"{self.quote_default(star.alias)}.*"
        else:
            return "*"

    def _build_count_sql(self, count: COUNT) -> str:
        count_sql = f"COUNT({self.build_expression_sql(count.expression)})"
        if count.field_alias:
            count_sql = f"{count_sql} AS {self.quote_default(count.field_alias)}"
        return count_sql

    def _build_distinct_sql(self, distinct: DISTINCT) -> str:
        expressions: list[SqlExpression] = (
            distinct.expression if isinstance(distinct.expression, list) else [distinct.expression]
        )
        field_expression_str = ", ".join([self.build_expression_sql(e) for e in expressions])
        return f"DISTINCT({field_expression_str})"

    def _build_sum_sql(self, sum: SUM) -> str:
        return f"SUM({self.build_expression_sql(sum.expression)})"

    def _build_is_null_sql(self, is_null: IS_NULL) -> str:
        return f"{self.build_expression_sql(is_null.expression)} IS NULL"

    def _build_is_not_null_sql(self, is_null: IS_NOT_NULL) -> str:
        return f"{self.build_expression_sql(is_null.expression)} IS NOT NULL"

    def _build_in_sql(self, in_: IN) -> str:
        list_expressions: str = ", ".join([self.build_expression_sql(element) for element in in_.list_expression])
        return f"{self.build_expression_sql(in_.expression)} IN ({list_expressions})"

    def _build_in_select_sql(self, in_select: IN_SELECT) -> str:
        nested_select: str = self.build_select_sql(
            select_elements=in_select.nested_select_elements, add_semicolon=False
        )
        nested_select: str = indent(nested_select, "    ")
        return f"{self.build_expression_sql(in_select.expression)} IN (\n{nested_select})"

    def _build_like_sql(self, like: LIKE) -> str:
        return f"{self.build_expression_sql(like.left)} LIKE {self.build_expression_sql(like.right)}"

    def _build_exists_sql(self, exists: EXISTS) -> str:
        nested_select: str = self.build_select_sql(select_elements=exists.nested_select_elements, add_semicolon=False)
        nested_select: str = indent(nested_select, "    ")
        return f"EXISTS (\n{nested_select})"

    def _build_not_like_sql(self, not_like: NOT_LIKE) -> str:
        return f"{self.build_expression_sql(not_like.left)} NOT LIKE {self.build_expression_sql(not_like.right)}"

    def _build_regex_like_sql(self, matches: REGEX_LIKE) -> str:
        expression: str = self.build_expression_sql(matches.expression)
        return f"REGEXP_LIKE({expression}, '{matches.regex_pattern}')"

    def _build_lower_sql(self, lower: LOWER) -> str:
        return f"LOWER({self.build_expression_sql(lower.expression)})"

    def _build_length_sql(self, length: LENGTH) -> str:
        return f"LENGTH({self.build_expression_sql(length.expression)})"

    def _build_max_sql(self, max: MAX) -> str:
        return f"MAX({self.build_expression_sql(max.expression)})"

    def _build_min_sql(self, min: MIN) -> str:
        return f"MIN({self.build_expression_sql(min.expression)})"

    def _build_average_sql(self, average: AVERAGE) -> str:
        return f"AVG({self.build_expression_sql(average.expression)})"

    def _build_coalesce_sql(self, coalesce: COALESCE) -> str:
        args: str = ", ".join([self.build_expression_sql(expression) for expression in coalesce.args])
        return f"COALESCE({args})"

    def _build_cast_sql(self, cast: CAST) -> str:
        to_type_text: str = (
            self.get_data_source_data_type_name_for_soda_data_type_name(cast.to_type)
            if isinstance(cast.to_type, SodaDataTypeName)
            else cast.to_type
        )
        return f"CAST({self.build_expression_sql(cast.expression)} AS {to_type_text})"

    def _build_case_when_sql(self, case_when: CASE_WHEN) -> str:
        return (
            f"CASE WHEN {self.build_expression_sql(case_when.condition)} "
            + f"THEN {self.build_expression_sql(case_when.if_expression)} "
            + (f"ELSE {self.build_expression_sql(case_when.else_expression)} " if case_when.else_expression else "")
            + "END"
        )

    def _build_order_by_lines(self, select_elements: list) -> list[str]:
        order_by_clauses: list[str] = []
        for select_element in select_elements:
            if isinstance(select_element, ORDER_BY_ASC) or isinstance(select_element, ORDER_BY_DESC):
                expression = select_element.expression
                direction: str = " ASC" if isinstance(select_element, ORDER_BY_ASC) else " DESC"
                order_by_clauses.append(f"{self.build_expression_sql(expression)}{direction}")
        if order_by_clauses:
            order_by_text: str = ", ".join(order_by_clauses)
            return [f"ORDER BY {order_by_text}"]
        else:
            return []

    def _build_limit_line(self, select_elements: list) -> Optional[str]:
        for select_element in select_elements:
            if isinstance(select_element, LIMIT):
                return self._build_limit_sql(select_element)

        return None

    def _build_offset_line(self, select_elements: list) -> Optional[str]:
        for select_element in select_elements:
            if isinstance(select_element, OFFSET):
                return self._build_offset_sql(select_element)

        return None

    def _build_ordinal_position_sql(self, ordinal_position: ORDINAL_POSITION) -> str:
        return "ORDINAL_POSITION"

    def _build_limit_sql(self, limit_element: LIMIT) -> str:
        return f"LIMIT {limit_element.limit}"

    def _build_offset_sql(self, offset_element: OFFSET) -> str:
        return f"OFFSET {offset_element.offset}"

    def supports_function(self, function: str) -> bool:
        return function in [
            "avg",
            "avg_length",
            "max",
            "min",
            "max_length",
            "min_length",
            "sum",
            "avglength",
            "maxlength",
            "minlength",
        ]

    def get_function_expression(self, function: str, arg: SqlExpression) -> SqlExpression:
        if function in ["avglength", "avg_length"]:
            return AVERAGE(LENGTH(arg))
        elif function in ["maxlength", "max_length"]:
            return MAX(LENGTH(arg))
        elif function in ["minlength", "min_length"]:
            return MIN(LENGTH(arg))
        elif function == "max":
            return MAX(arg)
        elif function == "min":
            return MIN(arg)
        elif function == "avg":
            return AVERAGE(arg)
        elif function == "sum":
            return SUM(arg)
        return FUNCTION(name=function, args=[arg])

    def _build_tuple_sql(self, tuple: TUPLE) -> str:
        elements: str = ", ".join(self.build_expression_sql(e) for e in tuple.expressions)
        return f"({elements})"

    def get_soda_null_string_value(self) -> str:
        return "__SODA_NULL__"

    def _build_concat_sql(self, concat: CONCAT) -> str:
        elements: str = ", ".join(self.build_expression_sql(e) for e in concat.expressions)
        return f"CONCAT({elements})"

    def _build_concat_ws_sql(self, concat_ws: CONCAT_WS) -> str:
        elements: str = ", ".join(self.build_expression_sql(e) for e in concat_ws.expressions)
        return f"CONCAT_WS({concat_ws.separator}, {elements})"

    def _build_string_hash_sql(self, string_hash: STRING_HASH) -> str:
        return f"MD5({self.build_expression_sql(string_hash.expression)})"

    def _build_combined_hash_sql(self, combined_hash: COMBINED_HASH) -> str:
        """Convert a set of columns into a unique hashed string which can be used as a key."""

        def format_expr(e: SqlExpression) -> SqlExpression:
            """Convert expression to a string, and replace nulls with a predefined string."""
            return COALESCE([CAST(e, to_type=SodaDataTypeName.VARCHAR), LITERAL(self.get_soda_null_string_value())])

        formatted_expressions: list[SqlExpression] = [format_expr(e) for e in combined_hash.expressions]
        if len(formatted_expressions) == 1:
            string_to_hash = formatted_expressions[0]
        else:
            string_to_hash = CONCAT_WS(separator="'||'", expressions=formatted_expressions)
        return self.build_expression_sql(STRING_HASH(string_to_hash))

    def _build_sample_sql(self, sampler_type: SamplerType, sample_size: Number) -> str:
        raise NotImplementedError("Sampling not implemented for this dialect")

    def information_schema_namespace_elements(self, data_source_namespace: DataSourceNamespace) -> list[str]:
        """
        The prefixes / namespace of the information schema for a given dataset prefix / namespace
        """
        database_name: str | None = data_source_namespace.get_database_for_metadata_query()
        if database_name:
            return [database_name, self.schema_information_schema()]
        else:
            return [self.schema_information_schema()]

    def schema_information_schema(self) -> str | None:
        """
        Name of the schema that has the metadata
        """
        return self.default_casify("information_schema")

    def table_tables(self) -> str:
        """
        Name of the table that has the table information in the metadata
        """
        return self.default_casify("tables")

    def table_columns(self) -> str:
        """
        Name of the table that has the columns information in the metadata.
        Purpose of this method is to allow specific data source to override.
        """
        return self.default_casify("columns")

    def table_schemata(self) -> str:
        """
        Name of the table that has the schema information in the metadata.
        Purpose of this method is to allow specific data source to override.
        """
        return self.default_casify("schemata")

    def column_schemata_catalog_name(self) -> str:
        """
        Name of the column that has the database/catalog information in the schemata metadata table.
        Purpose of this method is to allow specific data source to override.
        """
        return self.default_casify("catalog_name")

    def column_schema_name(self) -> str:
        """
        Name of the column that has the schema name in the schemata metadata table.
        Purpose of this method is to allow specific data source to override.
        """
        return self.default_casify("schema_name")

    def column_table_catalog(self) -> str:
        """
        Name of the column that has the database information in the tables metadata table
        """
        return self.default_casify("table_catalog")

    def column_table_schema(self) -> str:
        """
        Name of the column that has the schema information in the tables metadata table
        """
        return self.default_casify("table_schema")

    def column_table_name(self) -> str:
        """
        Name of the column that has the table name in the tables metadata table
        """
        return self.default_casify("table_name")

    def column_table_type(self) -> str:
        """
        Name of the column that has the table type in the tables metadata table
        """
        return self.default_casify("table_type")

    def convert_table_type_to_enum(self, table_type: str) -> TableType:
        if table_type == "BASE TABLE":
            return TableType.TABLE
        elif table_type == "VIEW":
            return TableType.VIEW
        elif table_type == "MATERIALIZED VIEW":
            return TableType.MATERIALIZED_VIEW
        else:
            # Default to TABLE if the table type is not recognized (so we're backwards compatible with existing code)
            logger.warning(f"Invalid table type: {table_type}, defaulting to TABLE")
            return TableType.TABLE

    def column_column_name(self) -> str:
        """
        Name of the column that has the column name in the tables metadata table.
        Purpose of this method is to allow specific data source to override.
        """
        return self.default_casify("column_name")

    def column_data_type(self) -> str:
        """
        Name of the column that has the data type in the columns metadata table.
        Purpose of this method is to allow specific data source to override.
        """
        return self.default_casify("data_type")

    def column_data_type_max_length(self) -> Optional[str]:
        """
        Name or definition of the column that has the max data type length in the columns metadata table.
        Purpose of this method is to allow specific data source to override.
        """
        return self.default_casify("character_maximum_length")

    def supports_data_type_character_maximum_length(self) -> bool:
        return True

    def column_data_type_numeric_precision(self) -> Optional[str]:
        """
        Name or definition of the column that has the max data type length in the columns metadata table.
        Purpose of this method is to allow specific data source to override.
        """
        return self.default_casify("numeric_precision")

    def supports_data_type_numeric_precision(self) -> bool:
        return True

    def column_data_type_numeric_scale(self) -> Optional[str]:
        """
        Name or definition of the column that has the max data type length in the columns metadata table.
        Purpose of this method is to allow specific data source to override.
        """
        return self.default_casify("numeric_scale")

    def supports_data_type_numeric_scale(self) -> bool:
        return True

    def column_data_type_datetime_precision(self) -> Optional[str]:
        """
        Name or definition of the column that has the max data type length in the columns metadata table.
        Purpose of this method is to allow specific data source to override.
        """
        return self.default_casify("datetime_precision")

    def supports_data_type_datetime_precision(self) -> bool:
        return True

    def supports_datetime_microseconds(self) -> bool:
        return True

    def default_casify(self, identifier: str) -> str:
        return identifier

    def metadata_casify(self, identifier: str) -> str:
        """Define case for metadata identifiers if needed."""
        return identifier

    def create_table_casify_qualified_name(self, qualified_name: str) -> str:
        # For data sources that default to a specific case for the table name (e.g. Snowflake prefers uppercase), this method can be overridden to casify the qualified table name.
        return qualified_name  # By default, we don't casify the qualified table name when creating a table as these are quoted

    # Very lightweight dialect-specific interpretation of dataset prefixes.
    def get_database_prefix_index(self) -> int | None:
        return 0

    # Very lightweight dialect-specific interpretation of dataset prefixes.
    def get_schema_prefix_index(self) -> int | None:
        return 1

    def sql_expr_timestamp_with_tz_literal(self, datetime_in_iso8601: str) -> str:
        """Convert to a SQL representation of a timestamp with timezone.

        By default this will return the standard SQL timestamp representation but may be overridden.
        We may wish to add some logic to detect timezones in datetime and return the appropriate representation.
        For now it's up to the user to decide which representation to use.
        """
        return self.sql_expr_timestamp_literal(datetime_in_iso8601)

    def sql_expr_timestamp_literal(self, datetime_in_iso8601: str) -> str:
        return f"timestamp '{datetime_in_iso8601}'"

    def sql_expr_timestamp_truncate_day(self, timestamp_literal: str) -> str:
        return f"date_trunc('day', {timestamp_literal})"

    def sql_expr_timestamp_add_day(self, timestamp_literal: str) -> str:
        return f"{timestamp_literal} + interval '1 day'"

    def quote_column(self, column_name: str) -> str:
        return self.quote_default(column_name)

    def format_metadata_data_type(self, data_type: str) -> str:
        """Allows processing data type string result from metadata column query if needed (Oracle uses this)."""
        return data_type

    def supports_regex_advanced(self) -> bool:
        return True  # Default to true, but specific dialects can override to false

    def encode_string_for_sql(self, string: str) -> str:
        """This escapes values that contain newlines correctly."""
        return string.encode("unicode_escape").decode("utf-8")

    def get_max_table_name_length(self) -> int:
        return 63

    def get_max_sql_statement_length(self) -> int:
        # What is the maximum query length of common analytical databases?
        # ChatGPT said:
        # Here are the maximum query lengths for some common analytical databases:
        # PostgreSQL: 1 GB
        # MySQL: 1 MB (configurable via max_allowed_packet)
        # SQL Server: 65,536 bytes (approximately 65 KB)
        # Oracle: 64 KB (depends on SQL string encoding)
        # Snowflake: 1 MB
        # BigQuery: No documented limit on query size, but practical limits on complexity and performance.
        return 63 * 1024 * 1024

    def get_preferred_number_of_rows_for_insert(self) -> int:
        # 10000 is the default number of rows for insert in a single SQL statement that is sent to the data source.
        # Please overwrite this for each data source as needed.
        # Currently, we set this value relateively low so the memory consumption of the Python library is not too high.
        return 10000

    def is_quoted(self, identifier: str) -> bool:
        return identifier.startswith(self.DEFAULT_QUOTE_CHAR) and identifier.endswith(self.DEFAULT_QUOTE_CHAR)

    def default_numeric_precision(self) -> Optional[int]:
        return None

    def default_numeric_scale(self) -> Optional[int]:
        return None

    def get_sql_data_type_class(self) -> type:
        return SqlDataType

    def supports_case_sensitive_column_names(self) -> bool:
        return True

    def apply_sampling(
        self,
        sql: str,
        sampler_limit: Number,
        sampler_type: SamplerType,
    ) -> str:
        return apply_sampling_to_sql(
            sql=sql,
            sampler_limit=sampler_limit,
            sampler_type=sampler_type,
            read_dialect=self.data_source_impl.type_name,
            write_dialect=self.data_source_impl.type_name,
        )

    ########################################################
    # Metadata columns query
    ########################################################

    def build_columns_metadata_from_clause(self, table_namespace: DataSourceNamespace) -> FROM:
        """
        Builds the FROM clause for querying column metadata.
        Purpose of this method is to allow specific data source to override.
        """
        return FROM(self.table_columns()).IN(self.information_schema_namespace_elements(table_namespace))

    def build_columns_metadata_query_str(self, table_namespace: DataSourceNamespace, table_name: str) -> str:
        """
        Builds the full SQL query to query table names from the data source metadata.
        """

        database_name: str | None = table_namespace.get_database_for_metadata_query()
        schema_name: str = table_namespace.get_schema_for_metadata_query()

        return self.build_select_sql(
            [
                SELECT(
                    [
                        self.column_column_name(),
                        self.column_data_type(),
                        *(
                            [self.column_data_type_max_length()]
                            if self.supports_data_type_character_maximum_length() and self.column_data_type_max_length()
                            else []
                        ),
                        *(
                            [self.column_data_type_numeric_precision()]
                            if self.supports_data_type_numeric_precision() and self.column_data_type_numeric_precision()
                            else []
                        ),
                        *(
                            [self.column_data_type_numeric_scale()]
                            if self.supports_data_type_numeric_scale() and self.column_data_type_numeric_scale()
                            else []
                        ),
                        *(
                            [self.column_data_type_datetime_precision()]
                            if self.supports_data_type_datetime_precision()
                            and self.column_data_type_datetime_precision()
                            else []
                        ),
                    ]
                ),
                self.build_columns_metadata_from_clause(table_namespace),
                WHERE(
                    AND(
                        [
                            *(
                                [EQ(self.column_table_catalog(), LITERAL(self.metadata_casify(database_name)))]
                                if database_name
                                else []
                            ),
                            EQ(self.column_table_schema(), LITERAL(self.metadata_casify(schema_name))),
                            EQ(self.column_table_name(), LITERAL(self.metadata_casify(table_name))),
                        ]
                    )
                ),
                ORDER_BY_ASC(ORDINAL_POSITION()),
            ]
        )

    def build_schemas_metadata_query_str(
        self, table_namespace: Optional[DataSourceNamespace] = None, filter_on_schema_name: Optional[str] = None
    ) -> str:
        """
        Builds the full SQL query to query schema names from the data source metadata.
        Optionally filters by database/catalog if provided in table_namespace.

        If the filter_on_schema_name is provided, it will only return the schema name that matches the filter. This is useful to check if a schema exists.
        """
        database_name: str | None = table_namespace.get_database_for_metadata_query() if table_namespace else None

        information_schema_namespace_elements = self.information_schema_namespace_elements(table_namespace)

        select_elements = [
            SELECT([self.column_schema_name()]),
            FROM(self.table_schemata()).IN(information_schema_namespace_elements),
        ]

        and_elements = [
            EQ(self.column_schemata_catalog_name(), LITERAL(self.metadata_casify(database_name))),
        ]

        if filter_on_schema_name:
            and_elements.append(EQ(self.column_schema_name(), LITERAL(self.metadata_casify(filter_on_schema_name))))

        if database_name:
            select_elements.append(WHERE(AND(and_elements)))

        return self.build_select_sql(select_elements)

    def extract_column_index(self, column_name: str, columns: list[Tuple[Any, ...]]) -> int:
        column_names = [c[0] for c in columns]
        return column_names.index(column_name)

    def extract_data_type_name(self, row: Tuple[Any, ...], columns: list[Tuple[Any, ...]]) -> str:
        return row[1]

    def extract_character_maximum_length(self, row: Tuple[Any, ...], columns: list[Tuple[Any, ...]]) -> Optional[int]:
        """Extract character maximum length from column metadata.  Typically this is just the value of a specific column."""
        data_type_name: str = self.extract_data_type_name(row, columns)
        if not self.data_type_has_parameter_character_maximum_length(data_type_name):
            return None
        if self.supports_data_type_character_maximum_length() and self.column_data_type_max_length():
            col_index = self.extract_column_index(self.column_data_type_max_length(), columns)
            return row[col_index]
        return None

    def extract_numeric_precision(self, row: Tuple[Any, ...], columns: list[Tuple[Any, ...]]) -> Optional[int]:
        """Extract numeric precision from column metadata.  Typically this is just the value of a specific column."""
        data_type_name: str = self.extract_data_type_name(row, columns)
        if not self.data_type_has_parameter_numeric_precision(data_type_name):
            return None

        if self.supports_data_type_numeric_precision() and self.column_data_type_numeric_precision():
            col_index = self.extract_column_index(self.column_data_type_numeric_precision(), columns)
            return row[col_index]
        return None

    def extract_numeric_scale(self, row: Tuple[Any, ...], columns: list[Tuple[Any, ...]]) -> Optional[int]:
        """Extract numeric scale from column metadata.  Typically this is just the value of a specific column."""
        data_type_name: str = self.extract_data_type_name(row, columns)
        if not self.data_type_has_parameter_numeric_scale(data_type_name):
            return None

        if self.supports_data_type_numeric_scale() and self.column_data_type_numeric_scale():
            col_index = self.extract_column_index(self.column_data_type_numeric_scale(), columns)
            return row[col_index]
        return None

    def extract_datetime_precision(self, row: Tuple[Any, ...], columns: list[Tuple[Any, ...]]) -> Optional[int]:
        """Extract datetime precision from column metadata.  Typically this is just the value of a specific column."""
        data_type_name: str = self.extract_data_type_name(row, columns)
        if not self.data_type_has_parameter_datetime_precision(data_type_name):
            return None

        if self.supports_data_type_datetime_precision() and self.column_data_type_datetime_precision():
            col_index = self.extract_column_index(self.column_data_type_datetime_precision(), columns)
            return row[col_index]
        return None

    def build_column_metadatas_from_query_result(self, query_result: QueryResult) -> list[ColumnMetadata]:
        column_metadatas: list[ColumnMetadata] = []
        for row in query_result.rows:
            column_name: str = row[0]
            data_type_name: str = self.format_metadata_data_type(self.extract_data_type_name(row, query_result.columns))
            character_maximum_length: Optional[int] = self.extract_character_maximum_length(row, query_result.columns)
            numeric_precision: Optional[int] = self.extract_numeric_precision(row, query_result.columns)
            numeric_scale: Optional[int] = self.extract_numeric_scale(row, query_result.columns)
            datetime_precision: Optional[int] = self.extract_datetime_precision(row, query_result.columns)

            column_metadatas.append(
                ColumnMetadata(
                    column_name=column_name,
                    sql_data_type=self.get_sql_data_type_class()(
                        name=data_type_name,
                        character_maximum_length=character_maximum_length,
                        numeric_precision=numeric_precision,
                        numeric_scale=numeric_scale,
                        datetime_precision=datetime_precision,
                    ),
                )
            )
        return column_metadatas

    def get_column_name(self, column: Any) -> str:
        """Extract the column name from an element in cursor.description"""
        if type(column) == tuple:
            return column[0]
        try:
            return column.name
        except:
            raise NotImplementedError(f"Unknown column type: {type(column)}")

    def get_column_index(self, column_name: str, columns: list[Any]) -> int:
        """Extract the column index from an element in cursor.description"""
        col_names = [self.get_column_name(c) for c in columns]
        return col_names.index(column_name)

    def convert_str_to_datetime(self, datetime_str: str) -> datetime:
        return convert_str_to_datetime(datetime_str)

    def convert_datetime_to_str(self, datetime: datetime) -> str:
        return convert_datetime_to_str(datetime)

    def supports_views(
        self,
    ) -> bool:  # Default to True, but can be overridden by specific data sources if they don't support views (Dremio)
        return True

    def supports_materialized_views(
        self,
    ) -> bool:  # Default to False, so that we roll out support for materialized views explicitly
        return False



import sys
import traceback
from typing import Optional

from soda_core.common.dataset_identifier import DatasetIdentifier


class SodaCoreException(Exception):
    """Base class for all data source connection exceptions."""

    def __init__(self, message: str, *args: object) -> None:
        super().__init__(message, *args)
        self.message = message


class InvalidArgumentException(SodaCoreException):
    """Indicates an invalid argument was passed to a function or method."""


class SodaCloudAuthenticationFailedException(SodaCoreException):
    """Indicates the authentication to Soda Cloud failed."""


class InvalidSodaCloudConfigurationException(SodaCoreException):
    """Indicates missing required keys in the Soda Cloud configuration file."""


class InvalidDataSourceConfigurationException(SodaCoreException):
    """Indicates the data source configuration is invalid."""


class DataSourceConnectionException(SodaCoreException):
    """Base class for all data source connection exceptions."""


class InvalidContractException(SodaCoreException):
    """Base class for all invalid contract exceptions."""


class ExtensionException(SodaCoreException):
    """Indicates that the extensions are not installed."""


class FailedContractSkeletonGenerationException(SodaCoreException):
    """Indicates that the contract skeleton generation failed."""


class InvalidRegexException(InvalidContractException):
    """Indicates the regex is invalid."""

    def __init__(self, sql: str):
        super().__init__(f"Invalid regex found in SQL '{sql}'")

    @classmethod
    def should_raise(cls, exception: Exception, sql: str) -> Optional["InvalidRegexException"]:
        if (
            hasattr(exception, "args")
            and exception.args[0] == "invalid regular expression: quantifier operand invalid\n"
        ):
            return cls(sql)
        return None


class InvalidDatasetQualifiedNameException(InvalidContractException):
    """Indicates the `dataset` property of the contract is not a valid Dataset Qualified Name"""


class YamlParserException(SodaCoreException):
    """Indicates an error occurred while parsing a YAML file."""

    def __init__(self, message: str, location: Optional[str] = None):
        message_with_location = f"{message}, in {location}" if location else message
        super().__init__(message_with_location)


class ContractParserException(YamlParserException):
    """Indicates an error occurred while parsing a contract."""


class SodaCloudException(SodaCoreException):
    """Base class for all SodaCloud related exceptions."""


class ContractNotFoundException(SodaCloudException):
    """Indicates the contract was not found in Soda Cloud."""

    def __init__(self, dataset_identifier: DatasetIdentifier):
        super().__init__(
            f"No data contract found for dataset '{str(dataset_identifier)}' in Soda Cloud. "
            "Please publish a contract for this dataset in Soda Cloud before proceeding."
        )


class DataSourceNotFoundException(SodaCloudException):
    """Indicates the data source was not found in Soda Cloud."""

    def __init__(self, dataset_identifier: DatasetIdentifier):
        super().__init__(
            f"Data source '{dataset_identifier.data_source_name}' is unknown in Soda Cloud. "
            "Please verify the data source name or configure it in Soda Cloud."
        )


class DatasetNotFoundException(SodaCloudException):
    """Indicates the dataset was not found in Soda Cloud."""

    def __init__(self, dataset_identifier: DatasetIdentifier):
        super().__init__(
            f"Dataset '{dataset_identifier.dataset_name}' is unknown in Soda Cloud. "
            "Please verify the dataset name or configure it in Soda Cloud."
        )


def get_exception_stacktrace(exception) -> Optional[str]:
    if isinstance(exception, BaseException):
        if sys.version_info < (3, 10):
            return "".join(
                traceback.format_exception(etype=type(exception), value=exception, tb=exception.__traceback__)
            )
        return "".join(traceback.format_exception(exception))
    return None



from __future__ import annotations

import contextlib
import logging
import os
from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import Any, Callable, Optional

from soda_core.common.data_source_results import (
    QueryResult,
    QueryResultIterator,
    UpdateResult,
)
from soda_core.common.logging_constants import soda_logger

logger: logging.Logger = soda_logger

from tabulate import tabulate


class DataSourceConnection(ABC):
    MAX_CHARS_PER_STRING = int(os.environ.get("SODA_DEBUG_PRINT_VALUE_MAX_CHARS", 256))
    MAX_ROWS = int(os.environ.get("SODA_DEBUG_PRINT_RESULT_MAX_ROWS", 20))
    MAX_CHARS_PER_SQL = int(os.environ.get("SODA_DEBUG_PRINT_SQL_MAX_CHARS", 1024))

    def __init__(self, name: str, connection_properties: dict, connection: Optional[object] = None):
        self.name: str = name
        self.connection_properties: dict = connection_properties
        self.connection: Optional[object] = connection

        # Auto-open on creation if no connection already supplied. See DataSource.open_connection()
        self.open_connection()

    def __enter__(self):
        pass

    @abstractmethod
    def _create_connection(self, connection_yaml_dict: dict) -> object:
        """
        self.connection_yaml_dict is provided as a parameter for convenience.
        Returns a new DBAPI connection based on the provided connection properties.
        The returned value will be saved in self.connection. See open() for details.
        Exceptions do not need to be handled.  They will be handled by the calling method.
        """

    def open_connection(self) -> None:
        """
        Ensures that an open connection is available.
        After this method ends, the open connection must have the database_name and optionally
        the schema_name as context as far as these are applicable for the specific data source type.
        Potentially closes and re-opens the self.connection if the existing connection has a different database
        or schema context compared to the given database_name and schema_name for the next contract.
        """
        if self.connection is None:
            try:
                logger.debug(f"'{self.name}' connection properties: {self.connection_properties}")
                self.connection = self._create_connection(self.connection_properties)
            except Exception as e:
                logger.error(msg=f"Could not connect to '{self.name}': {e}", exc_info=True)

    def close_connection(self) -> None:
        """
        Closes te connection. This method will not throw any exceptions.
        Check errors with has_errors or assert_no_errors.
        """
        if self.connection:
            try:
                # noinspection PyUnresolvedReferences
                self.connection.close()
            except Exception as e:
                logger.warning(msg=f"Could not close the DBAPI connection", exc_info=True)
            finally:
                self.connection = None

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_connection()

    def execute_query(self, sql: str, log_query: bool = True) -> QueryResult:
        # noinspection PyUnresolvedReferences
        cursor = self.connection.cursor()
        try:
            if log_query:
                logger.debug(
                    f"SQL query fetchall in datasource {self.name} (first {self.MAX_CHARS_PER_SQL} chars): \n{self.truncate_sql(sql)}"
                )

            cursor.execute(sql)
            rows = cursor.fetchall()
            formatted_rows = self.format_rows(rows)
            truncated_rows = self.truncate_rows(formatted_rows)
            headers = [self._execute_query_get_result_row_column_name(c) for c in cursor.description]
            # The tabulate can crash if the rows contain non-ASCII characters.
            # This is purely for debugging/logging purposes, so we can try/catch this.
            try:
                table_text: str = tabulate(
                    truncated_rows,
                    headers=headers,
                    tablefmt="github",
                )
            except UnicodeDecodeError as e:
                logger.debug(f"Error formatting rows. These may contain non-ASCII characters. {e}")
                table_text = "Error formatting rows. These may contain non-ASCII characters."

            logger.debug(
                f"SQL query result (max {self.MAX_ROWS} rows, {self.MAX_CHARS_PER_STRING} chars per string):\n{table_text}"
            )
            return QueryResult(rows=formatted_rows, columns=cursor.description)
        finally:
            cursor.close()

    def format_rows(self, rows: list[tuple]) -> list[tuple]:
        return rows

    def truncate_sql(self, sql: str) -> str:
        """Truncate large strings in sql to a reasonable length."""
        if len(sql) > (self.MAX_CHARS_PER_SQL - 3):
            return sql[: self.MAX_CHARS_PER_SQL - 3] + "..."
        return sql

    def truncate_rows(self, rows: list[tuple]) -> list[tuple]:
        """Truncate large strings in rows to a reasonable length, and return only the first n rows."""

        def truncate_cell(cell: Any) -> Any:
            if isinstance(cell, str) and len(cell) > (self.MAX_CHARS_PER_STRING - 3):
                return cell[: self.MAX_CHARS_PER_STRING - 3] + "..."
            return cell

        if len(rows) > self.MAX_ROWS:
            rows = rows[: self.MAX_ROWS]
        rows = [tuple(truncate_cell(cell) for cell in row) for row in rows]

        return rows

    def _execute_query_get_result_row_column_name(self, column) -> str:
        return column.name

    def execute_update(self, sql: str, log_query: bool = True) -> UpdateResult:
        # noinspection PyUnresolvedReferences
        cursor = self.connection.cursor()
        try:
            if log_query:
                logger.debug(f"SQL update (first {self.MAX_CHARS_PER_SQL} chars): \n{self.truncate_sql(sql)}")
            return self._cursor_execute_update_and_commit(cursor, sql)

        finally:
            cursor.close()

    def _cursor_execute_update_and_commit(self, cursor: Any, sql: str):
        updates = cursor.execute(sql)
        self.commit()
        return updates

    def execute_query_one_by_one(
        self,
        sql: str,
        row_callback: Callable[[tuple, tuple[tuple]], None],
        log_query: bool = True,
        row_limit: Optional[int] = None,
    ) -> tuple[tuple]:
        """
        usage: execute_query_one_by_one("SELECT ...", lambda row, description: your_handle_row(row))
        Returns the description of the query.
        """
        # noinspection PyUnresolvedReferences
        cursor = self.connection.cursor()
        try:
            if log_query:
                logger.debug(f"SQL query fetch one-by-one:\n{sql}")
            cursor.execute(sql)

            description: tuple[tuple] = cursor.description

            rows_processed: int = 0

            row = cursor.fetchone()
            while row and (row_limit is None or rows_processed < row_limit):
                rows_processed += 1
                row_callback(row, description)
                row = cursor.fetchone()

        finally:
            cursor.close()
        return description

    @contextlib.contextmanager
    def execute_query_iterate(self, sql: str, log_query: bool = True) -> Iterator[QueryResultIterator]:
        cursor = self.connection.cursor()
        if log_query:
            logger.debug(f"SQL query iterate:\n{sql}")
        try:
            cursor.execute(sql)
            yield QueryResultIterator(cursor)
        finally:
            cursor.close()

    def commit(self) -> None:
        # noinspection PyUnresolvedReferences
        self.connection.commit()

    def rollback(self) -> None:
        # noinspection PyUnresolvedReferences
        self.connection.rollback()

    def disable_close_connection(self) -> None:
        self.close_connection_enabled = False

    def enable_close_connection(self) -> None:
        self.close_connection_enabled = True



from __future__ import annotations

import signal
import sys
import traceback
from argparse import ArgumentParser, _SubParsersAction
from typing import Dict, List, Optional, Union

from soda_core.__version__ import SODA_CORE_VERSION
from soda_core.cli.exit_codes import ExitCode
from soda_core.cli.handlers.contract import (
    handle_fetch_contract,
    handle_publish_contract,
    handle_test_contract,
    handle_verify_contract,
)
from soda_core.cli.handlers.data_source import (
    handle_create_data_source,
    handle_test_data_source,
)
from soda_core.cli.handlers.request import (
    handle_fetch_proposal,
    handle_push_proposal,
    handle_transition_request,
)
from soda_core.cli.handlers.soda_cloud import (
    handle_create_soda_cloud,
    handle_test_soda_cloud,
)
from soda_core.common.logging_configuration import configure_logging
from soda_core.common.logging_constants import soda_logger
from soda_core.contracts.contract_request import RequestStatus
from soda_core.telemetry.soda_telemetry import SodaTelemetry
from soda_core.telemetry.soda_tracer import soda_trace

CLOUD_CONFIG_PATH_HELP = "A Soda Cloud configuration file path."
FILE_PATH_HELP = "The path to the file to be created. (directories will be created if needed)"
REQUEST_NUMBER_HELP = "The Contract Request number"

# Legacy v3 commands that are no longer supported in v4
LEGACY_V3_COMMANDS = frozenset(
    [
        "scan",
        "scan_status",
        "ingest",
        "test_connection",
        "simulate_anomaly_detection",
    ]
)

soda_telemetry = SodaTelemetry()


@soda_trace
def execute() -> None:
    try:
        print(r"  __|  _ \|  \   \\")
        print(r"\__ \ (   |   | _ \\")
        print(r"____/\___/___/_/  _\\ CLI v%s" % SODA_CORE_VERSION)

        signal.signal(signal.SIGINT, handle_ctrl_c)

        handle_legacy_commands()

        args = cli_parser.parse_args()

        soda_telemetry.ingest_cli_arguments(vars(args))

        if len(sys.argv) == 1:
            cli_parser.print_help()
            sys.exit(ExitCode.LOG_ERRORS)

        verbose = args.verbose if hasattr(args, "verbose") else False
        _configure_logging(verbose)

        if not hasattr(args, "handler_func"):
            soda_logger.error(f"No handler found for resource '{args.resource}' and command '{args.command}'")
            exit_with_code(ExitCode.LOG_ERRORS)

        args.handler_func(args)

    except Exception as e:
        traceback.print_exc()
        exit_with_code(ExitCode.LOG_ERRORS)


def _configure_logging(verbose: bool) -> None:
    """
    Purpose of this method is to enable override in test environment
    """
    configure_logging(verbose=verbose)


def handle_legacy_commands():
    if len(sys.argv) > 1 and sys.argv[1] in LEGACY_V3_COMMANDS:
        print("Soda v3 commands are not supported (Soda v4 was run)")
        print("Please run Soda v3, or execute only Soda v4 commands.")
        print("See https://docs.soda.io/soda-v4/reference/cli-reference for more information.")

        exit_with_code(ExitCode.LOG_ERRORS)


def create_cli_parser() -> ArgumentParser:
    parser = ArgumentParser(
        prog="soda",
        epilog="Run 'soda {resource} {command} -h' for help on a specific command",
    )
    resource_parsers = parser.add_subparsers(dest="resource", help="Available Soda resources")

    _setup_contract_resource(resource_parsers)
    _setup_data_source_resource(resource_parsers)
    _setup_soda_cloud_resource(resource_parsers)
    _setup_contract_request_resource(resource_parsers)

    return parser


def _setup_contract_resource(resource_parsers) -> None:
    contract_parser = resource_parsers.add_parser("contract", help="Contract commands")
    contract_subparsers = contract_parser.add_subparsers(dest="command", help="Contract commands")

    _setup_contract_verify_command(contract_subparsers)
    _setup_contract_publish_command(contract_subparsers)
    _setup_contract_test_command(contract_subparsers)
    _setup_contract_fetch_command(contract_subparsers)


def _setup_contract_verify_command(contract_parsers) -> None:
    verify_parser = contract_parsers.add_parser("verify", help="Verify a contract")

    verify_parser.add_argument(
        "-c",
        "--contract",
        type=str,
        help="Contract file path to verify. Use this to work with local contracts.",
    )
    verify_parser.add_argument(
        "-d",
        "--dataset",
        type=str,
        help="Name of dataset to verify. Use this to work with remote contracts present in Soda Cloud.",
    )

    verify_parser.add_argument(
        "-ds", "--data-source", type=str, help="The data source configuration file.", nargs="+", default=[]
    )
    verify_parser.add_argument("-sc", "--soda-cloud", type=str, help=CLOUD_CONFIG_PATH_HELP)
    verify_parser.add_argument(
        "--set",
        action="append",
        type=str,
        help="Set variable values to be used in the contract with format '--set <variable_name>=<variable_value>'.",
    )
    verify_parser.add_argument(
        "-a",
        "--use-agent",
        const=True,
        action="store_const",
        default=False,
        help="Executes contract verification on Soda Agent instead of locally in this library.",
    )
    verify_parser.add_argument(
        "-btm",
        "--blocking-timeout-in-minutes",
        type=int,
        default=60,
        help="Max time in minutes that the CLI should wait for the contract "
        "verification to complete on Soda Agent.  Default is 60 minutes.",
    )
    verify_parser.add_argument(
        "-p",
        "--publish",
        const=True,
        action="store_const",
        default=False,
        help="Send the verification results to Soda Cloud.",
    )
    verify_parser.add_argument(
        "-v",
        "--verbose",
        const=True,
        action="store_const",
        default=False,
        help="Show more detailed logs on the console.",
    )
    verify_parser.add_argument(
        "-cp",
        "--check-paths",
        type=str,
        nargs="*",
        help="One or more check paths to run in a contract.",
    )
    # TODO: move into extensions
    verify_parser.add_argument(
        "-dw",
        "--diagnostics-warehouse",
        type=str,
        nargs="?",
        help="Specify the path to the diagnostics warehouse configuration file. ",
    )

    def handle(args):
        contract_file_path = args.contract
        dataset_identifier = args.dataset
        data_source_file_paths = args.data_source
        soda_cloud_file_path = args.soda_cloud
        variables = _parse_variables(args.set)
        if variables is None:
            exit_with_code(ExitCode.LOG_ERRORS)
        publish = args.publish
        verbose = args.verbose
        use_agent = args.use_agent
        blocking_timeout_in_minutes = args.blocking_timeout_in_minutes
        check_paths = args.check_paths
        diagnostics_warehouse_file_path = args.diagnostics_warehouse

        exit_code = handle_verify_contract(
            contract_file_path,
            dataset_identifier,
            data_source_file_paths,
            soda_cloud_file_path,
            variables,
            publish,
            verbose,
            use_agent,
            blocking_timeout_in_minutes,
            check_paths,
            diagnostics_warehouse_file_path,
        )

        exit_with_code(exit_code)

    verify_parser.set_defaults(handler_func=handle)


def _parse_variables(variables: Optional[List[str]]) -> Optional[Dict[str, str]]:
    if not variables:
        return {}

    result = {}
    for variable in variables:
        if "=" not in variable:
            soda_logger.error(f"Variable {variable} is incorrectly formatted. Please use the format KEY=VALUE")
            return None
        key, value = variable.split("=", 1)
        key = key.strip()
        value = value.strip()
        value = _parse_variable_value(key=key, value=value)
        if not key or not value:
            soda_logger.error(
                f"Incorrectly formatted variable '{variable}', key or value is empty. Please use the format KEY=VALUE"
            )
            return None
        result[key] = value
    return result


def _parse_variable_value(key: str, value: str) -> Union[str, float, int]:
    # Try to parse the value as an integer first
    # Note: we don't use the .isnumeric() or isdigit() methods here, as they are not reliable for floats, negative numbers...
    try:
        value = int(value)
        soda_logger.info(f"Variable {key} is a number, parsed as int: {value}")
        return value
    except ValueError:
        pass
    # Then try to parse it as a float
    try:
        value = float(value)
        soda_logger.info(f"Variable {key} is a number, parsed as float: {value}")
        return value
    except ValueError:
        pass
    # If it's not a number, return the string
    return value


def _setup_contract_publish_command(contract_parsers) -> None:
    publish_parser = contract_parsers.add_parser("publish", help="Publish a contract")
    publish_parser.add_argument("-c", "--contract", type=str, help="Contract file path to publish.")

    publish_parser.add_argument(
        "-sc",
        "--soda-cloud",
        type=str,
        help=CLOUD_CONFIG_PATH_HELP,
        required=True,
    )

    publish_parser.add_argument(
        "-v",
        "--verbose",
        const=True,
        action="store_const",
        default=False,
        help="Show more detailed logs on the console.",
    )

    def handle(args):
        contract_file_path = args.contract
        soda_cloud_file_path = args.soda_cloud
        exit_code = handle_publish_contract(contract_file_path, soda_cloud_file_path)
        exit_with_code(exit_code)

    publish_parser.set_defaults(handler_func=handle)


def _setup_contract_test_command(contract_parsers) -> None:
    test_contract_parser = contract_parsers.add_parser(name="test", help="Test a contract syntax without executing it")
    test_contract_parser.add_argument("-c", "--contract", type=str, help="Contract file path to test.")

    test_contract_parser.add_argument(
        "-v",
        "--verbose",
        const=True,
        action="store_const",
        default=False,
        help="Show more detailed logs on the console.",
    )

    def handle(args):
        contract_file_path = args.contract

        exit_code = handle_test_contract(contract_file_path, {})
        exit_with_code(exit_code)

    test_contract_parser.set_defaults(handler_func=handle)


def _setup_contract_fetch_command(contract_parsers) -> None:
    fetch_parser = contract_parsers.add_parser("fetch", help="Pull a contract")
    fetch_parser.add_argument(
        "-d",
        "--dataset",
        type=str,
        nargs="+",
        help="Fully qualified names of datasets whose cloud contracts you wish to fetch.",
    )
    fetch_parser.add_argument(
        "-sc",
        "--soda-cloud",
        type=str,
        help=CLOUD_CONFIG_PATH_HELP,
        required=True,
    )
    fetch_parser.add_argument(
        "-f",
        "--file",
        type=str,
        nargs="+",
        help="The path(s) to the contract files to be created or updated. (directories will be created if needed)",
    )
    fetch_parser.add_argument(
        "-v",
        "--verbose",
        const=True,
        action="store_const",
        default=False,
        help="Show more detailed logs on the console.",
    )

    def handle(args):
        contract_file_paths = args.file
        soda_cloud_file_path = args.soda_cloud
        dataset_identifiers = args.dataset

        exit_code = handle_fetch_contract(contract_file_paths, dataset_identifiers, soda_cloud_file_path)
        exit_with_code(exit_code)

    fetch_parser.set_defaults(handler_func=handle)


def _setup_data_source_resource(resource_parsers) -> None:
    data_source_parser = resource_parsers.add_parser("data-source", help="Data source commands")
    data_source_subparsers = data_source_parser.add_subparsers(dest="command", help="Data source commands")

    _setup_data_source_create_command(data_source_subparsers)
    _setup_data_source_test_command(data_source_subparsers)


def _setup_data_source_create_command(data_source_parsers) -> None:
    create_data_source_parser = data_source_parsers.add_parser(
        name="create", help="Create a data source YAML configuration file"
    )
    create_data_source_parser.add_argument(
        "-f",
        "--file",
        type=str,
        help=FILE_PATH_HELP,
    )
    create_data_source_parser.add_argument(
        "-t", "--type", type=str, default="postgres", help="Type of the data source.  Eg postgres"
    )

    create_data_source_parser.add_argument(
        "-v",
        "--verbose",
        const=True,
        action="store_const",
        default=False,
        help="Show more detailed logs on the console.",
    )

    def handle(args):
        data_source_file_path = args.file
        data_source_type = args.type

        exit_code = handle_create_data_source(data_source_file_path, data_source_type)
        exit_with_code(exit_code)

    create_data_source_parser.set_defaults(handler_func=handle)


def _setup_data_source_test_command(data_source_parsers) -> None:
    test_parser = data_source_parsers.add_parser("test", help="Test a data source connection")
    test_parser.add_argument("-ds", "--data-source", type=str, help="The name of a configured data source to test.")

    test_parser.add_argument(
        "-v",
        "--verbose",
        const=True,
        action="store_const",
        default=False,
        help="Show more detailed logs on the console.",
    )

    def handle(args):
        data_source_file_path = args.data_source

        exit_code = handle_test_data_source(data_source_file_path)
        exit_with_code(exit_code)

    test_parser.set_defaults(handler_func=handle)


def _setup_soda_cloud_resource(resource_parsers) -> None:
    soda_cloud_parser = resource_parsers.add_parser("cloud", help="Soda Cloud commands")
    soda_cloud_subparsers = soda_cloud_parser.add_subparsers(dest="command", help="Soda Cloud commands")

    _setup_soda_cloud_create_command(soda_cloud_subparsers)
    _setup_soda_cloud_test_command(soda_cloud_subparsers)


def _setup_soda_cloud_create_command(soda_cloud_parsers) -> None:
    create_soda_cloud_parser = soda_cloud_parsers.add_parser(
        name="create", help="Create a Soda Cloud YAML configuration file"
    )
    create_soda_cloud_parser.add_argument(
        "-f",
        "--file",
        type=str,
        help=FILE_PATH_HELP,
    )

    create_soda_cloud_parser.add_argument(
        "-v",
        "--verbose",
        const=True,
        action="store_const",
        default=False,
        help="Show more detailed logs on the console.",
    )

    def handle(args):
        soda_cloud_file_path = args.file
        exit_code = handle_create_soda_cloud(soda_cloud_file_path)
        exit_with_code(exit_code)

    create_soda_cloud_parser.set_defaults(handler_func=handle)


def _setup_soda_cloud_test_command(soda_cloud_parsers) -> None:
    test_soda_cloud_parser = soda_cloud_parsers.add_parser("test", help="Test the Soda Cloud connection")
    test_soda_cloud_parser.add_argument("-sc", "--soda-cloud", type=str, help=CLOUD_CONFIG_PATH_HELP)

    test_soda_cloud_parser.add_argument(
        "-v",
        "--verbose",
        const=True,
        action="store_const",
        default=False,
        help="Show more detailed logs on the console.",
    )

    def handle(args):
        soda_cloud_file_path = args.soda_cloud
        exit_code = handle_test_soda_cloud(soda_cloud_file_path)
        exit_with_code(exit_code)

    test_soda_cloud_parser.set_defaults(handler_func=handle)


def _setup_contract_request_resource(resource_parsers) -> None:
    contract_request_parser = resource_parsers.add_parser("request", help="Contract request commands")
    contract_request_subparsers = contract_request_parser.add_subparsers(
        dest="command", help="Contract request commands"
    )

    _setup_contract_request_fetch_proposal_command(contract_request_subparsers)
    _setup_contract_request_push_proposal_command(contract_request_subparsers)
    _setup_contract_request_transition_command(contract_request_subparsers)


def _setup_contract_request_fetch_proposal_command(contract_request_parsers: ArgumentParser) -> None:
    fetch_proposal_parser = contract_request_parsers.add_parser("fetch", help="Fetch a proposal from Soda Cloud")
    fetch_proposal_parser.add_argument(
        "-sc",
        "--soda-cloud",
        type=str,
        required=True,
        help=CLOUD_CONFIG_PATH_HELP,
    )

    fetch_proposal_parser.add_argument(
        "-r",
        "--request",
        type=int,
        required=True,
        help=REQUEST_NUMBER_HELP,
    )

    fetch_proposal_parser.add_argument(
        "-f",
        "--file",
        type=str,
        required=True,
        help=FILE_PATH_HELP,
    )

    fetch_proposal_parser.add_argument(
        "-p",
        "--proposal",
        type=int,
        help="The Proposal number (optional). The latest proposal will be fetched if not provided.",
    )

    def handle(args):
        soda_cloud_file_path = args.soda_cloud
        request_number = args.request
        output_file_path = args.file
        proposal_number = args.proposal

        exit_code = handle_fetch_proposal(
            soda_cloud_file_path=soda_cloud_file_path,
            request_number=request_number,
            output_file_path=output_file_path,
            proposal_number=proposal_number,
        )

        exit_with_code(exit_code)

    fetch_proposal_parser.set_defaults(handler_func=handle)


def _setup_contract_request_push_proposal_command(contract_request_parsers: ArgumentParser) -> None:
    push_proposal_parser = contract_request_parsers.add_parser("push", help="Push a proposal to Soda Cloud")

    push_proposal_parser.add_argument(
        "-sc",
        "--soda-cloud",
        type=str,
        required=True,
        help=CLOUD_CONFIG_PATH_HELP,
    )

    push_proposal_parser.add_argument(
        "-f",
        "--file",
        type=str,
        required=True,
        help=FILE_PATH_HELP,
    )

    push_proposal_parser.add_argument(
        "-r",
        "--request",
        type=int,
        required=True,
        help=REQUEST_NUMBER_HELP,
    )

    push_proposal_parser.add_argument(
        "-m",
        "--message",
        type=str,
        help="The message to be pushed to Soda Cloud (optional)",
    )

    def handle(args) -> None:
        soda_cloud_file_path = args.soda_cloud
        file_path = args.file
        request_number = args.request
        message = args.message

        exit_code = handle_push_proposal(
            soda_cloud_file_path=soda_cloud_file_path,
            file_path=file_path,
            request_number=request_number,
            message=message,
        )

        exit_with_code(exit_code)

    push_proposal_parser.set_defaults(handler_func=handle)


def _setup_contract_request_transition_command(contract_request_parsers: ArgumentParser) -> None:
    transition_request_parser = contract_request_parsers.add_parser(
        "transition", help="Transition a request's status on Soda Cloud"
    )

    transition_request_parser.add_argument(
        "-sc",
        "--soda-cloud",
        type=str,
        required=True,
        help=CLOUD_CONFIG_PATH_HELP,
    )

    transition_request_parser.add_argument(
        "-r",
        "--request",
        type=int,
        required=True,
        help=REQUEST_NUMBER_HELP,
    )

    transition_request_parser.add_argument(
        "-s",
        "--status",
        type=str,
        required=True,
        choices=[status.value for status in RequestStatus],
        help="The status to transition the request to",
    )

    def handle(args) -> None:
        soda_cloud_file_path = args.soda_cloud
        request_number = args.request
        status = RequestStatus(args.status)

        exit_code = handle_transition_request(
            soda_cloud_file_path=soda_cloud_file_path,
            request_number=request_number,
            status=status,
        )
        exit_with_code(exit_code)

    transition_request_parser.set_defaults(handler_func=handle)


def exit_with_code(exit_code: int):
    soda_logger.debug(f"Exiting with code {exit_code}")
    soda_telemetry.set_attribute("cli__exit_code", exit_code)
    exit(exit_code)


def handle_ctrl_c(self, sig, frame):
    soda_logger.info("")
    soda_logger.info(f"CTRL+C detected")
    exit_with_code(ExitCode.LOG_ERRORS)


def get_or_create_resource_parser(
    root_parser: ArgumentParser, resource_name: str, help_str: Optional[str] = None
) -> ArgumentParser:
    resource_subparsers = _get_or_create_subparsers(root_parser, "resource")
    if resource_name in resource_subparsers.choices:
        return resource_subparsers.choices[resource_name]

    return resource_subparsers.add_parser(resource_name, help=help_str)


def get_or_create_command_parser(
    root_parser: ArgumentParser, resource_name: str, command_name: str, help_str: Optional[str] = None
) -> ArgumentParser:
    resource_parser = get_or_create_resource_parser(root_parser, resource_name)
    command_subparsers = _get_or_create_subparsers(resource_parser, "command")

    if command_name in command_subparsers.choices:
        return command_subparsers.choices[command_name]

    return command_subparsers.add_parser(command_name, help=help_str)


def _get_or_create_subparsers(parser: ArgumentParser, dest: str, help_str: Optional[str] = None) -> _SubParsersAction:
    for action in parser._actions:
        if isinstance(action, _SubParsersAction) and action.dest == dest:
            return action

    return parser.add_subparsers(dest=dest, help=help_str)


cli_parser = create_cli_parser()


if __name__ == "__main__":
    execute()



import json
from collections.abc import Sequence

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult


class MemorySpanExporter(SpanExporter):
    """Implementation of :class:`SpanExporter` that saves spans in memory.

    This class can be used for diagnostic purposes, multi-threaded scenarios etc.
    """

    __instance = None
    __spans = []

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
            cls.__instance._initialize()
        return cls.__instance

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        for span in spans:
            self.__spans.append(span)
        return SpanExportResult.SUCCESS

    def reset(self):
        self.__spans = []

    @property
    def spans(self) -> list[ReadableSpan]:
        return self.__spans

    @property
    def span_dicts(self) -> list[dict]:
        return [json.loads(span.to_json()) for span in self.spans]



from soda_core.common import soda_cloud_dto, sql_ast


def map_sampler_type_from_dto(sampler_type_dto: soda_cloud_dto.SamplerType) -> sql_ast.SamplerType:
    if sampler_type_dto is None:
        return None
    elif sampler_type_dto == soda_cloud_dto.SamplerType.ABSOLUTE_LIMIT:
        return sql_ast.SamplerType.ABSOLUTE_LIMIT
    else:
        raise ValueError(f"Unsupported sampler type DTO: {sampler_type_dto}")



from __future__ import annotations

from soda_core.common.sql_dialect import *
from soda_core.contracts.contract_verification import CheckOutcome, CheckResult
from soda_core.contracts.impl.check_types.missing_check_yaml import MissingCheckYaml
from soda_core.contracts.impl.check_types.row_count_check import RowCountMetricImpl
from soda_core.contracts.impl.contract_verification_impl import (
    AggregationMetricImpl,
    CheckImpl,
    CheckParser,
    ColumnImpl,
    ContractImpl,
    DerivedPercentageMetricImpl,
    MeasurementValues,
    MetricImpl,
    MissingAndValidityCheckImpl,
    ThresholdImpl,
    ThresholdType,
)


class MissingCheckParser(CheckParser):
    def get_check_type_names(self) -> list[str]:
        return ["missing"]

    def parse_check(
        self,
        contract_impl: ContractImpl,
        column_impl: Optional[ColumnImpl],
        check_yaml: MissingCheckYaml,
    ) -> Optional[CheckImpl]:
        return MissingCheckImpl(
            contract_impl=contract_impl,
            column_impl=column_impl,
            check_yaml=check_yaml,
        )


class MissingCheckImpl(MissingAndValidityCheckImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        column_impl: ColumnImpl,
        check_yaml: MissingCheckYaml,
    ):
        super().__init__(
            contract_impl=contract_impl,
            column_impl=column_impl,
            check_yaml=check_yaml,
        )
        self.threshold = ThresholdImpl.create(
            threshold_yaml=check_yaml.threshold,
            default_threshold=ThresholdImpl(type=ThresholdType.SINGLE_COMPARATOR, must_be=0),
        )
        self.metric_name = "missing_percent" if check_yaml.metric == "percent" else "missing_count"

    def setup_metrics(
        self,
        contract_impl: ContractImpl,
        column_impl: ColumnImpl,
        check_yaml: MissingCheckYaml,
    ):
        self.missing_count_metric_impl = self._resolve_metric(
            MissingCountMetricImpl(contract_impl=contract_impl, column_impl=column_impl, check_impl=self)
        )

        self.row_count_metric_impl: MetricImpl = self._resolve_metric(
            RowCountMetricImpl(contract_impl=contract_impl, check_impl=self)
        )

        self.missing_percent_metric_impl: MetricImpl = self.contract_impl.metrics_resolver.resolve_metric(
            DerivedPercentageMetricImpl(
                metric_type="missing_percent",
                fraction_metric_impl=self.missing_count_metric_impl,
                total_metric_impl=self.row_count_metric_impl,
            )
        )

    def evaluate(self, measurement_values: MeasurementValues) -> CheckResult:
        outcome: CheckOutcome = CheckOutcome.NOT_EVALUATED

        missing_count: int = measurement_values.get_value(self.missing_count_metric_impl)
        row_count: int = measurement_values.get_value(self.row_count_metric_impl)
        missing_percent: float = measurement_values.get_value(self.missing_percent_metric_impl)

        diagnostic_metric_values: dict[str, float] = {
            "missing_count": missing_count,
            "missing_percent": missing_percent,
            "check_rows_tested": row_count,
            "dataset_rows_tested": self.contract_impl.dataset_rows_tested,
        }

        threshold_value: Optional[Number] = missing_percent if self.metric_name == "missing_percent" else missing_count

        outcome = self.evaluate_threshold(threshold_value)

        return CheckResult(
            check=self._build_check_info(),
            outcome=outcome,
            threshold_value=threshold_value,
            diagnostic_metric_values=diagnostic_metric_values,
        )

    def get_threshold_metric_impl(self) -> Optional[MetricImpl]:
        return self.missing_count_metric_impl


class MissingCountMetricImpl(AggregationMetricImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        column_impl: ColumnImpl,
        check_impl: MissingAndValidityCheckImpl,
    ):
        super().__init__(
            contract_impl=contract_impl,
            column_impl=column_impl,
            metric_type=check_impl.type,
            check_filter=check_impl.check_yaml.filter,
            missing_and_validity=check_impl.missing_and_validity,
        )

    def sql_expression(self) -> SqlExpression:
        return SUM(CASE_WHEN(self.sql_condition_expression(), LITERAL(1)))

    def sql_condition_expression(self) -> SqlExpression:
        column_name: str = self.column_impl.column_yaml.name
        not_missing_and_invalid_expr = self.missing_and_validity.is_missing_expr(column_name)
        return (
            not_missing_and_invalid_expr
            if not self.check_filter
            else AND([SqlExpressionStr(self.check_filter), not_missing_and_invalid_expr])
        )

    def convert_db_value(self, value) -> int:
        # Note: expression SUM(CASE WHEN "id" IS NULL THEN 1 ELSE 0 END) gives NULL / None as a result if
        # there are no rows
        return int(value) if value is not None else 0
